header
{
    package com.unitesk.java5;

    import java.io.File;
    import java.util.List;
    import java.util.EnumSet;
    import com.unitesk.atp.text.location.*;
    import com.unitesk.atp.messages.*;
    import com.unitesk.atpj5.tree.tool.antlr.ArrayListFactory;
    import com.unitesk.atpj5.tree.tool.antlr.ListFactory;
    import com.unitesk.atpj5.tree.tool.antlr.NullListFactory;
    import com.unitesk.java5.Java5.*;

    import static com.unitesk.java5.TokenUtils.getPosition;
}

class Java5Parser extends Parser;
options
{
    k = 2;
    importVocab = Java5Lexer;
//    codeGenBitsetTestThreshold=999;
}

{
    /**
      * Class containing boolean flag to be used in primary rule
      * for flagging situations: whether qualified identifier is
      * matched at primary expression start or not.
     */
    protected class IsPrimaryPrefixQID {
        public boolean flag = false;
    }

    /**
      * Method sets the flag to "true" if a primary expression
      * begins with a qualified identifier and return "true"
      * to allow its use as an expression inside of semantic
      * predicate.
      */
    protected boolean turnOnIsPrefixQID(IsPrimaryPrefixQID f) {
        f.flag = true;
        return true;
    }

    /**
      * Method sets the flag to "false" if a primary expression
      * doesn't begin with a qualified identifier and return "true"
      * to allow its use as an expression inside of semantic
      * predicate.
      */
    protected boolean turnOffIsPrefixQID(IsPrimaryPrefixQID f) {
        f.flag = false;
        return true;
    }

    /**
      * Counts the number of LTHN ("<") seen in the typeArguments production.
      * It is used in semantic predicates to ensure we have seen
      * enough closing '>' characters; which actually may have been
      * either GTHN (">"), RSHIFT (">>") or URSHIFT (">>>") tokens.
     */
    protected int ltCounter = 0;

    protected final Java5Factory nodeCreator = new Java5NodeFactory();
    protected final ListFactory arrayListFactory = new ArrayListFactory();

    protected final Java5Factory nullCreator = new Java5NullFactory();
    protected final ListFactory nullListFactory = new NullListFactory();

    boolean createTree = true;

    protected Java5Factory creator;
    protected ListFactory listFactory;

    protected File file;
    protected MessageReceiverAdapter mbox;
    protected Java5StartPositionVisitor startVisitor;
    protected Java5EndPositionVisitor endVisitor;

    protected JavaDocFilter javadocFilter;

    public void setFile( File file )
    {
        this.file = file;
    }

    public void setMessageBox( MessageReceiverAdapter mbox )
    {
        this.mbox = mbox;
    }

    public void setStartVisitor( Java5StartPositionVisitor startVisitor )
    {
        this.startVisitor = startVisitor;
    }

    public void setEndVisitor( Java5EndPositionVisitor endVisitor )
    {
        this.endVisitor = endVisitor;
    }

    public void setJavaDocFilter( JavaDocFilter f )
    {
        this.javadocFilter = f;
    }
    
    protected void createTree( boolean create )
    {
        if( create )
        {
            creator =  nodeCreator;
            listFactory = arrayListFactory;
        } else {
            creator =  nullCreator;
            listFactory = nullListFactory;
        }
        createTree = create;
    }

    protected void stopCreateTree()
    {
        createTree( false );
    }

    private String tokenText( Token t )
    {
        String text = t.getText();
        if( text != null )
        {
            return  "'" + text + "'";
        } else {
            eofFound = true;
            return "EOF";
        }
    }

    private boolean eofFound = false;

    public void reportError( RecognitionException ex )
    {
        stopCreateTree();
        
        if( eofFound ) return;

        if( ex instanceof NoViableAltException )
        {
            NoViableAltException e = (NoViableAltException)ex;
            String s = tokenText( e.token );
            if( !eofFound )
            {
	            try
	            {
	                s += ' ' + tokenText( LT(2) );
	            } 
	            catch( TokenStreamException tse ) {} 
	        }
            mbox.error( file
                      , new Position( ex.getLine()
                                    , ex.getColumn()
                                    )
                      , new Message.UnexpectedToken( s )
                      );
        } else
        if(    ex instanceof MismatchedTokenException
            && ((MismatchedTokenException)ex).mismatchType == MismatchedTokenException.TOKEN
          )
        {
            MismatchedTokenException e = (MismatchedTokenException)ex;
            mbox.error( file
                      , new Position( ex.getLine()
                                    , ex.getColumn()
                                    )
                      , new Message.MismatchedToken( _tokenNames[e.expecting], tokenText( e.token ) )
                      );
        } else {
            mbox.error( file
                      , new Position( ex.getLine()
                                    , ex.getColumn()
                                    )
                      , new Message.SyntaxError( ex.getMessage() )
                      );
        }
    }
    
    private void removeInappropriates( List<Token> optModifierList, java.util.Map<Modifier, EnumSet<Modifier>> inappropriatesMethodModifiers )
    {
       	if ( optModifierList == null )
       		return;
       		
		java.util.Map<Modifier, Token> savedModifiers = TokenUtils.getModifiersAsMap( optModifierList );
        java.util.Map<Modifier, EnumSet<Modifier>> errorModifierPairs = 
        	Modifier.removeInappropriates( 
        			optModifierList,
        			inappropriatesMethodModifiers );
        java.util.Set<Token> forDelete = new java.util.HashSet<Token>();
        for( Modifier modifier : errorModifierPairs.keySet() )
        {
    		Token mToken = savedModifiers.get( modifier );
        		
        	EnumSet<Modifier> errModifiers = errorModifierPairs.get( modifier );
        	for( Modifier m : errModifiers )
        	{
                mbox.error( file
                          , getPosition( mToken )
                          , new Message.IllegalModifiersCombination( modifier.toString(), m.toString() )
                          );
                forDelete.add( savedModifiers.get( m ) );
        	}
        }
        optModifierList.removeAll( forDelete ); 
	}    
	
	protected static final java.util.Map<Modifier, EnumSet<Modifier>> 
		methodDeclarationInappropriatesModifiers = new java.util.HashMap<Modifier, EnumSet<Modifier>>();
		
	static {
			methodDeclarationInappropriatesModifiers.put(Modifier.ABSTRACT, EnumSet.of(Modifier.STATIC, Modifier.STRICTFP, Modifier.PRIVATE));
			methodDeclarationInappropriatesModifiers.put(Modifier.STATIC, EnumSet.of(Modifier.ABSTRACT));
			methodDeclarationInappropriatesModifiers.put(Modifier.STRICTFP, EnumSet.of(Modifier.ABSTRACT));
			methodDeclarationInappropriatesModifiers.put(Modifier.PRIVATE, EnumSet.of(Modifier.ABSTRACT));
	}
	
}

//===== tree description ======================================================

/**
 *  Compilation unit - start rule.
 */
compilationUnit returns [CompilationUnit compilationUnit = null]
{
    PackageDeclaration optPackageDeclaration = null;
    List<ImportDeclaration> optImportDeclarationList = null;
    List<TypeDeclaration> optTypeDeclarationList = null;
}
:
    {
      createTree( true );
      eofFound = false;
    }
    // A compilation unit starts with an optional package definition
    ( ( optAnnotationList PACKAGE )=> optPackageDeclaration = packageDeclaration )?
    // Next we have a series of zero or more import statements
    optImportDeclarationList = optImportDeclarationList
    // And the last could be zero or more type declarations
    optTypeDeclarationList = optTypeDeclarationList
    (CTRL_Z)? EOF
    {
        compilationUnit = creator.createCompilationUnit( optPackageDeclaration
                                                       , optImportDeclarationList
                                                       , optTypeDeclarationList
                                                       );
    }
;

/**
 * Package declaration
 */
packageDeclaration returns [PackageDeclaration packageDeclaration = null]
{
    List<Annotation> optAnnotationList = null;
    QualifiedIdentifier name = null;
    Token startToken = null;
}
:
    { startToken = LT(1); }
    optAnnotationList = optAnnotationList
    PACKAGE
    name = qualifiedIdentifier
    endToken:SEMICOLON
    {
        packageDeclaration = creator.createPackageDeclaration
                                     ( getPosition( startToken )
                                     , getPosition( endToken )
                                     , optAnnotationList
                                     , name
                                     );
    }
;

/**
 * Zero or more import declarations
 */
optImportDeclarationList returns [List<ImportDeclaration> optImportDeclarationList = null]
{
    ImportDeclaration importDeclaration = null;
    boolean cTree = createTree;
}
:
    {
        createTree( true );
        optImportDeclarationList = listFactory.<ImportDeclaration>create();
    }
    ( importDeclaration = importDeclaration
      {
        listFactory.add( optImportDeclarationList, importDeclaration );
        createTree( true );
      }
    )*
    {
        createTree( cTree );
    }
;

/**
 * Import declaration
 */
importDeclaration returns [ImportDeclaration importDeclaration = null]
{
    QualifiedIdentifier name = null;
    boolean is_static = false;
    boolean is_on_demand = false;
}
:
    startToken:IMPORT
    ( STATIC { is_static = true; } )?
    name = qualifiedIdentifier
    ( DOT MULT { is_on_demand = true; } )?
    endToken:SEMICOLON
    {
        if( is_static )
        {
            if( is_on_demand )
            {
                importDeclaration
                    = creator.createStaticImportOnDemandDeclaration( getPosition( startToken ), getPosition( endToken ), name );
            } else {
                importDeclaration
                    = creator.createSingleStaticImportDeclaration( getPosition( startToken ), getPosition( endToken ), name );
            }
        } else {
            if( is_on_demand )
            {
                importDeclaration
                    = creator.createTypeImportOnDemandDeclaration( getPosition( startToken ), getPosition( endToken ), name );
            } else {
                importDeclaration
                    = creator.createSingleTypeImportDeclaration( getPosition( startToken ), getPosition( endToken ), name );
            }
        }
    }
;

/**
 * Zero or more top-level type declarations or semicolons
 */
optTypeDeclarationList returns [List<TypeDeclaration> optTypeDeclarationList = null]
{
    List<Annotation> optAnnotationList = null;
    List<Token> optModifierList = null;
    TypeDeclaration typeDeclaration = null;
    Token startToken = null;
    boolean cTree = createTree;
}
:
    {
      createTree( true );
      optTypeDeclarationList = listFactory.<TypeDeclaration>create();
    }
    (   {
            optAnnotationList = listFactory.<Annotation>create();
            optModifierList = listFactory.<Token>create();
            startToken = LT(1);
        }
        modifiers[optAnnotationList,optModifierList]
        typeDeclaration = typeDeclaration[startToken,optAnnotationList,optModifierList]
        {
          listFactory.add( optTypeDeclarationList, typeDeclaration );
          createTree( true );
        }
      | SEMICOLON // skip empty declaration
    )*
    {
        createTree( cTree );
    }
;

/**
 * Extention point to define extra categories of type declarations.
 */
typeDeclaration[ Token startToken
               , List<Annotation> optAnnotationList
               , List<Token> optModifierList
               ]
    returns [TypeDeclaration typeDeclaration = null]
:
    typeDeclaration = javaTypeDeclaration[startToken,optAnnotationList,optModifierList]
;

/**
 * Java type declaration
 */
javaTypeDeclaration[ Token startToken
                   , List<Annotation> optAnnotationList
                   , List<Token> optModifierList
                   ]
    returns [TypeDeclaration typeDecl = null]
:
       // inner class
       typeDecl = classDeclaration[startToken,optAnnotationList,optModifierList]
     | // inner enum
       typeDecl = enumDeclaration[startToken,optAnnotationList,optModifierList]
     | // inner interface
       typeDecl = interfaceDeclaration[startToken,optAnnotationList,optModifierList]
     | // inner annotation
       typeDecl = annotationTypeDeclaration[startToken,optAnnotationList,optModifierList]
;

/**
 * Java modifier. This rule match any possible modifier.
 * Context dependent conditions will be checked separately
 * to allow clear error messages "Modifier ... can't be used in ..."
 * instead of "Syntax error: unexpected token ..."
 */
javaModifier returns [Token modifier = null]
:
    { modifier = LT(1); }
    (   PUBLIC
      | PROTECTED
      | PRIVATE
      | ABSTRACT
      | STATIC
      | STRICTFP
      | FINAL
      | TRANSIENT
      | NATIVE
      | SYNCHRONIZED
      | VOLATILE
    )
;

/**
 * Extention point to define extra modifiers.
 */
modifier returns [Token modifier = null]
:
    modifier = javaModifier
;

/**
 * Optional sequence of modifiers and annotations.
 * @param  optAnnotationList   list to add annotations to
 * @param  optModifierList     list to add modifiers to
 * @return                     the first token of this sequence or null if it is empty
 */
modifiers[List<Annotation> optAnnotationList, List<Token> optModifierList]
{
    Token modifier = null;
    Annotation annotation = null;
    java.util.Map<String, Token> nameModifierMap = null;
    boolean cTree = createTree;
    java.util.List<Modifier> visitedModifiers = null;
    java.util.Set<Modifier> modifiersForDelete = null;
}
:
    {
    	nameModifierMap = new java.util.HashMap<String,Token>();
	    visitedModifiers = new java.util.ArrayList<Modifier>();
    	modifiersForDelete = new java.util.HashSet<Modifier>();
    	createTree( true );
    }
    ( // turn off warnings since the semantic check for
      // "@interface" solves the non-determinism
      options { generateAmbigWarnings=false; }:
        modifier = modifier
        {
            String m = modifier.getText();
            Modifier mod = TokenUtils.tokenToModifier( modifier );
            // mod == null means that it isn't java modifier (it's possible in language extentions)
            if( mod != null ) 
            {
	            if( nameModifierMap.containsKey( m ) )
	            {
		            // taking off duplicates (public public)
	                mbox.error( file
	                          , getPosition( modifier )
	                          , new Message.DuplicateModifier
	                                ( m, getPosition( nameModifierMap.get( m ) ) )
	                          );
	            } else if ( ! mod.corresponds( visitedModifiers ) )
	            {
		            // taking off inappropriate modifiers ( public private )
		            EnumSet<Modifier> currentInappropriates = EnumSet.copyOf( visitedModifiers );
		            currentInappropriates.retainAll( mod.getInappropriates() );
		            modifiersForDelete.addAll( currentInappropriates );
	                nameModifierMap.put( m, modifier );
	                visitedModifiers.add( mod );
	                mbox.error( file
	                          , getPosition( modifier )
	                          , new Message.IllegalModifiersCombination( m, currentInappropriates.iterator().next().toString() )
	                          );
	            } else {
	                nameModifierMap.put( m, modifier );
	                visitedModifiers.add( mod );
	                //listFactory.add( optModifierList, modifier );
			    	createTree( true );
	            }
	        }
        }
      | // Semantic check that "@interface" is not been matching
        // as it is not an annotation but annotation type declaration
        { LA(1) == AT && LA(2) != INTERFACE }?
        annotation = annotation
        {
            listFactory.add( optAnnotationList, annotation );
            createTree( true );
        }
    )*
    {
    	for( Modifier m1 : modifiersForDelete )
    		visitedModifiers.remove( m1 );
    		
    	if ( ! visitedModifiers.isEmpty() )
    	{
    		for( Modifier m2 : visitedModifiers )
    		{
    			listFactory.add( optModifierList, nameModifierMap.get( m2.toString() ) );
    		}
    	}
    		
        createTree( cTree );
    }
;

//=====  Qualification of type  ================================================

/**
 * Type or void
 */
optionalType[boolean bounds_allowed] returns [OptionalType optionalType = null]
:
      t:VOID { optionalType = creator.createVoidType( getPosition( t ) ); }
    | optionalType = type[bounds_allowed]
;

// A type name which is either a (possibly qualified and parameterized)
// class name or a primitive (built-in) type
// with possible brackets afterwards
// (which would make it an array type)
type[boolean bounds_allowed] returns [Type t = null]
{ int dim = 0; }
:
    (   t = qualifiedParameterizedIdentifier[bounds_allowed] // A non-primitive type name,
                                             // with possible type parameters,
                                             // or a type variable
      | t = primitiveType
    )
    ( // options{greedy=true;}: // match as many as possible
      LBRACKET endToken:RBRACKET { dim++; }
    )*

    { if( dim > 0 ) t = creator.createArrayType( getPosition( endToken ), t, dim ); }
;

parameterizedIdentifier[boolean bounds_allowed] returns [ParameterizedIdentifier paramIdent = null]
{
    Identifier ident = null;
    TypeArguments optTypeArguments = null;
}
:
    ident = identifier
    optTypeArguments = optTypeArguments[true, bounds_allowed]
    { paramIdent = creator.createParameterizedIdentifier( ident, optTypeArguments ); }
;

// A (possibly-qualified) (possibly-parameterized) java identifier.
// We start with the first IDENTIFIER (and possibly its arguments)
// and expand its name by adding dots and following IDENTIFIERS
// (and possibly their arguments)
qualifiedParameterizedIdentifier[boolean bounds_allowed] returns [QualifiedParameterizedIdentifier qpi = null]
{
    ParameterizedIdentifier paramIdent = null;
    List<ParameterizedIdentifier> paramIdentList = null;
    int currentLtLevel = 0;
    boolean cTree = createTree;
}
:
    {
        createTree( true );
        currentLtLevel = ltCounter;
        paramIdentList = listFactory.<ParameterizedIdentifier>create();
        createTree = cTree;
    }
    paramIdent = parameterizedIdentifier[bounds_allowed]
    {
        listFactory.add( paramIdentList, paramIdent );
        cTree = createTree;
    }
    ( options { greedy = true; }: // match as many as possible
      { createTree( true ); }
      { inputState.guessing != 0 || ltCounter == currentLtLevel }?
      DOT
      paramIdent = parameterizedIdentifier[bounds_allowed]
      {
		listFactory.add( paramIdentList, paramIdent );
	    createTree( true );
      }
    )*
    {
        qpi = creator.createQualifiedParameterizedIdentifier( paramIdentList );
        createTree( cTree );
    }
;

optTypeArguments[boolean wild_cards_allowed, boolean bounds_allowed] returns [TypeArguments typeArguments = null]
:
    ( typeArguments = typeArguments[wild_cards_allowed, bounds_allowed] )?
;

// Type arguments to a class or interface type
typeArguments[boolean wild_cards_allowed, boolean bounds_allowed] returns [TypeArguments typeArguments = null]
{
    int currentLtLevel = 0;
    ActualTypeArgument arg = null;
    List<ActualTypeArgument> type_args = null;
    Token endToken = null;
    Position endPosition = null;
    boolean cTree = createTree;
}
:
    {
        currentLtLevel = ltCounter;
    }
    startToken:LTHN
    {
        ltCounter++;
        createTree( true );
        type_args = listFactory.create();
        createTree( cTree );
    }
    arg = typeArgument[wild_cards_allowed, bounds_allowed]
    {
        listFactory.add(type_args, arg);
        cTree = createTree;
    }
    ( options { greedy = true; }: // match as many as possible
      { inputState.guessing != 0 || ltCounter == currentLtLevel + 1 }?
      COMMA
      { createTree( true ); }
      arg = typeArgument[wild_cards_allowed, bounds_allowed]
      {
        listFactory.add(type_args, arg);
      }
    )*
    {
        createTree( true );
    }
    (   // turn warning off since Antlr generates the right code,
        // plus we have our semantic predicate below
        options { generateAmbigWarnings = false; }:
        endToken = typeArgumentsOrParametersEnd
        { endPosition = getPosition( endToken ); }
      | { if( arg != null ) endPosition = endVisitor.getEndPosition( arg ); }

    )
    // make sure we have gobbled up enough '>' characters
    // if we are at the "top level" of nested typeArgument productions
    {
        createTree( cTree );
        
        if( currentLtLevel == 0 && ltCounter != currentLtLevel )
        {
            Token t = LT(1);
            mbox.error( file
                      , getPosition( t )
                      , new Message.MismatchedToken( "'>'", tokenText( t ) )
                      );
            stopCreateTree();
        }

        typeArguments = creator.createTypeArguments( getPosition( startToken ), endPosition, type_args );
    }
;

// Gobble up *some* amount of '>' characters, and counts how many
// they gobbled.
protected typeArgumentsOrParametersEnd returns [Token endToken = null]
:
    { endToken = LT(1); }
    (   GTHN    { ltCounter -= 1; }
      | RSHIFT  { ltCounter -= 2; }
      | URSHIFT { ltCounter -= 3; }
    )
;

// A generic type argument is a class type, a possibly bounded wildcard type
// or a built-in type array
typeArgument[boolean wild_cards_allowed, boolean bounds_allowed] returns [ActualTypeArgument arg = null]
{
    ReferenceType ref_type = null;
    WildcardActualTypeArgument wildcard_arg = null;
    Token t = new Token();
}
:
    ( 
       ( primitiveType ~LBRACKET) =>
       { t = LT(1); } 
        primitiveType
        {
        	mbox.error( 
        		file,
        		getPosition( t ),
        		new Message.ReferenceTypeExpectedButNotPrimitive()
        	);
        	
        	stopCreateTree();
        }
	|    	
        ref_type = referenceType[false]
        {
           arg = creator.createReferenceTypeActualTypeArgument(ref_type);
           t = LT(1);
           q = LT(1);
        }
      |
        ( q:QUESTION
        { wildcard_arg = creator.createWildcardActualTypeArgument( getPosition( q ) );
            t = LT(1);
            if( !wild_cards_allowed )
            {
                mbox.error( file
                          , getPosition( q )
                          , new Message.UnexpectedWildcard()
                          );
            }
        }
        )
    )

    ( ( EXTENDS | SUPER )=> ( arg = boundedWilcardType[q]
                             { if ( wildcard_arg == null && ! bounds_allowed )
                               { mbox.error( file
                                           , getPosition( t )
                                           , new Message.TypeBoundsAreForbidden()
                                           );
                                }
                             }
                             )
    )?

    {
        if( arg == null )
        arg = wildcard_arg;
    }
;

// A specialised form of type where built in types must be arrays
referenceType[boolean bounds_allowed] returns [ReferenceType rt = null]
{
    int dim = 0;
    int currentLtLevel = 0;
}
:
      { currentLtLevel = ltCounter; }
      // A non-primitive type name, with possible type parameters,
      // or a type variable
      rt = qualifiedParameterizedIdentifier[bounds_allowed]
      ( options { greedy=true; }: // match as many as possible
        { inputState.guessing != 0 || ltCounter == currentLtLevel }?
        LBRACKET endToken:RBRACKET { dim++; }
      )*
      { if( dim > 0 ) rt = creator.createArrayType( getPosition( endToken ), rt, dim ); }
    | // Array of primitive type
      rt = primitiveTypeArray
;

// A primitive type with brackets afterwards
primitiveTypeArray returns [ArrayType arrayType = null]
{
    PrimitiveType primitiveType = null;
    int dim = 0;
}
:
    primitiveType = primitiveType
    ( options { greedy=true; }: // match as many as possible
      LBRACKET endToken:RBRACKET { dim++; }
    )+

    { arrayType = creator.createArrayType( getPosition( endToken ), primitiveType, dim ); }
;

// Wildcard type indicating all types with constraint
boundedWilcardType[Token q] returns [WildcardActualTypeArgument wildcard_arg = null]
{ ReferenceType rt = null; }
:
      EXTENDS
      rt = referenceType[false]
      { wildcard_arg = creator.createExtendsWildcardActualTypeArgument( getPosition( q ), rt ); }
    | SUPER
      rt = referenceType[false]
      { wildcard_arg = creator.createSuperWildcardActualTypeArgument( getPosition( q ), rt ); }
;

// The primitive types.
primitiveType returns [PrimitiveType prim_type = null]
{
    NumericTypeKind kind = null;
    Token token = null;
}
:
    { token = LT(1); }
    (   BOOLEAN
        { prim_type = creator.createBooleanType( getPosition( token ) ); }
      | (   BYTE  { kind = NumericTypeKind.BYTE; }
          | SHORT { kind = NumericTypeKind.SHORT; }
          | INT   { kind = NumericTypeKind.INT; }
          | LONG  { kind = NumericTypeKind.LONG; }
          | CHAR  { kind = NumericTypeKind.CHAR; }
        )
        { prim_type = creator.createIntegralType( getPosition( token ), kind ); }
      | (   FLOAT  { kind = NumericTypeKind.FLOAT; }
          | DOUBLE { kind = NumericTypeKind.DOUBLE; }
        )
        { prim_type = creator.createFloatingPointType( getPosition( token ), kind ); }
    )
;

//=====  Class declaration  ====================================================

typeBody[Token startToken] returns [TypeBody typeBody = null]
{
    MemberDeclaration member = null;
    List<MemberDeclaration> optMemberList = null;
    boolean cTree = createTree;
}
:
    {
        createTree( true );
        optMemberList = listFactory.<MemberDeclaration>create();
    }
    (   member = memberDeclaration
        {
            listFactory.add( optMemberList, member );
            createTree( true );
        }
      | SEMICOLON
    )*
    {
        typeBody = creator.createTypeBody( getPosition( startToken ), getPosition( LT(1) ), optMemberList );
        createTree( cTree );
    }
;

/**
 * Class declaration.
 * @param optAnnotationList list of annotations for this class declaration
 * @param optModifierList   list of modifiers for this class declaration
 */
classDeclaration[ Token startToken
                , List<Annotation> optAnnotationList
                , List<Token> optModifierList
                ]
    returns [ClassDeclaration classDeclaration = null]
{
    Identifier name = null;
    TypeParameters optTypeParameters = null;
    Type optSourceSuperType = null;
    List<QualifiedParameterizedIdentifier> optInterfaceList = null;
    TypeBody typeBody = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    CLASS
    name = identifier
    optTypeParameters = optTypeParameters
    ( EXTENDS
//      optSuperClass = qualifiedParameterizedIdentifier[true]
      optSourceSuperType = type[true]
    )?
    {
    	cTree = createTree;
    	createTree( true );
    }
    (  IMPLEMENTS optInterfaceList = optInterfaceList
     |
        {
            optInterfaceList = listFactory.<QualifiedParameterizedIdentifier>create();
        }
    )
    {
    	createTree( cTree );
    }
    typeBodyStartToken:LBRACE
    typeBody = typeBody[typeBodyStartToken]
    {
    	Token endToken = LT(1);
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        classDeclaration = creator.createClassDeclaration
                                   ( getPosition( startToken )
                                   , getPosition( endToken )
                                   , typeBody
                                   , javadoc
                                   , optAnnotationList
                                   , modifiers
                                   , name
                                   , optTypeParameters
                                   , optSourceSuperType
                                   , optInterfaceList
                                   );
         if( optSourceSuperType != null )
         {
         	 if ( optSourceSuperType.key() == QualifiedParameterizedIdentifier.KEY )
	         {
	         	classDeclaration.setOptSuperClass( (QualifiedParameterizedIdentifier) optSourceSuperType ); 
	         }
	         else if ( optSourceSuperType.key() == ArrayType.KEY )
	         {
	         	mbox.error( file, startVisitor.getStartPosition( optSourceSuperType ), new Message.ArrayTypeAsSuperType() );
	         }
	         else
	         {
	         	mbox.error( file, startVisitor.getStartPosition( optSourceSuperType ), new Message.ReferenceTypeExpectedButNotPrimitive() );
	         }
	      }
    }
    RBRACE
;

optTypeParameters returns [TypeParameters optTypeParameters = null]
{
    int currentLtLevel = 0;
    TypeParameter param = null;
    List<TypeParameter> typeParameterList = null;
    Token endToken = null;
    Position endPosition = null;
    boolean cTree = createTree;
}
:
{
    createTree( true );
    currentLtLevel = ltCounter;
    typeParameterList = listFactory.<TypeParameter>create();
    createTree( cTree );
}
  ( startToken:LTHN
    { ltCounter++; }
    param = typeParameter
    {
        listFactory.add( typeParameterList, param);
        cTree = createTree;
    }
    (  { createTree( true ); }
      COMMA
      param = typeParameter
      { listFactory.add( typeParameterList, param ); }
    )*
    {
        createTree( cTree );
    }
    (   endToken = typeArgumentsOrParametersEnd
        { endPosition = getPosition( endToken ); }
      | { if( param != null ) endPosition = endVisitor.getEndPosition( param ); }
    )

    // make sure we have gobbled up enough '>' characters
    // if we are at the "top level" of nested typeArgument productions
    {
        if( currentLtLevel == 0 && ltCounter != currentLtLevel )
        {
            Token t = LT(1);
            mbox.error( file
                      , getPosition( t )
                      , new Message.MismatchedToken( "'>'", tokenText( t ) )
                      );
            stopCreateTree();
        }
    }
    { optTypeParameters = creator.createTypeParameters( getPosition( startToken ), endPosition, typeParameterList ); }
  )?
;

typeParameter returns [TypeParameter param = null]
{
    Identifier type_variable = null;
    QualifiedParameterizedIdentifier opt_class_or_interface = null;
    QualifiedParameterizedIdentifier bound_interface = null;
    List<QualifiedParameterizedIdentifier> opt_additional_bound_interfaces = null;
    boolean cTree = createTree;
}
:
{
    createTree( true );
    opt_additional_bound_interfaces = listFactory.<QualifiedParameterizedIdentifier>create();
    createTree( cTree );
}
    type_variable = identifier
    { cTree = createTree; }
    ( // I'm pretty sure Antlr generates the right thing here:???
      options { generateAmbigWarnings = false; }:
      EXTENDS
      opt_class_or_interface = qualifiedParameterizedIdentifier[false]
      ( { createTree( true ); }
        BIT_AND
        bound_interface = qualifiedParameterizedIdentifier[false]
        {
            listFactory.add(opt_additional_bound_interfaces, bound_interface);
        }
      )*
    )?
    
    {	createTree( cTree );

        param = creator.createTypeParameter(   type_variable
                                             , opt_class_or_interface
                                             , opt_additional_bound_interfaces
                                           );
    }
;

/**
 * One or more comma-separated interface names
 */
optInterfaceList returns [List<QualifiedParameterizedIdentifier> optInterfaceList = null]
{
    Type interfaceSourceName = null;
    boolean cTree = createTree;
}
:
	{ optInterfaceList = listFactory.<QualifiedParameterizedIdentifier>create(); }
    interfaceSourceName = type[false] 
    {
         if( interfaceSourceName != null )
         {
         	 if ( interfaceSourceName.key() == QualifiedParameterizedIdentifier.KEY )
	         {
					listFactory.add( optInterfaceList, (QualifiedParameterizedIdentifier)interfaceSourceName );
	         }
	         else if ( interfaceSourceName.key() == ArrayType.KEY )
	         {
	         	mbox.error( file, startVisitor.getStartPosition( interfaceSourceName ), new Message.ArrayTypeAsSuperInterface() );
	         }
	         else
	         {
	         	mbox.error( file, startVisitor.getStartPosition( interfaceSourceName ), new Message.ReferenceTypeExpectedButNotPrimitive() );
	         }
	      }
    
    
        cTree = createTree;
    }
    ( { createTree( true ); }
      COMMA
      interfaceSourceName = type[false]
      {
         if( interfaceSourceName != null )
         {
         	 if ( interfaceSourceName.key() == QualifiedParameterizedIdentifier.KEY )
	         {
					listFactory.add( optInterfaceList, (QualifiedParameterizedIdentifier)interfaceSourceName );
	         }
	         else if ( interfaceSourceName.key() == ArrayType.KEY )
	         {
	         	mbox.error( file, startVisitor.getStartPosition( interfaceSourceName ), new Message.ArrayTypeAsSuperInterface() );
	         }
	         else
	         {
	         	mbox.error( file, startVisitor.getStartPosition( interfaceSourceName ), new Message.ReferenceTypeExpectedButNotPrimitive() );
	         }
	      }
      }
    )*
    {
        createTree( cTree );
    }
;

//=====  Enum declaration  =====================================================

/**
 * Enum declaration.
 */
enumDeclaration[ Token startToken
               , List<Annotation> optAnnotationList
               , List<Token> optModifierList
               ]
    returns [EnumDeclaration enumDeclaration = null]
{
    Identifier name = null;
    List<QualifiedParameterizedIdentifier> optInterfaceList = null;
    EnumConstant constant = null;
    List<EnumConstant> optConstantList = null;
    TypeBody typeBody = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    ENUM
    name = identifier
    {
    	cTree = createTree;
    	createTree( true );
    }
    (  IMPLEMENTS optInterfaceList = optInterfaceList
     |
        {
            optInterfaceList = listFactory.<QualifiedParameterizedIdentifier>create();
        }
    )

    // This is the body of a enum. It can contain enum constants,
    // class fields and extra commas/semicolons.
    LBRACE
    {
    	createTree( true );
    	optConstantList = listFactory.<EnumConstant>create();
    }
    ( constant = enumConstant
      {
        listFactory.add( optConstantList, constant );
      }
      ( options { greedy = true; }:
      	{ createTree( true ); }
        COMMA
        constant = enumConstant
        { listFactory.add( optConstantList, constant ); }
      )*
    )?
    {
        createTree( cTree );
    }
    ( COMMA )?
    (   typeBodyStartToken:SEMICOLON
        typeBody = typeBody[typeBodyStartToken]
      | {
      Position position = getPosition( LT(1) );
            typeBody = creator.createTypeBody( position, position, listFactory.<MemberDeclaration>create() );
        }
    )
    {
	    Token endToken = LT(1);
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        enumDeclaration = creator.createEnumDeclaration
                                  ( getPosition( startToken )
                                  , getPosition( endToken )
                                  , typeBody
                                  , javadoc
                                  , optAnnotationList
                                  , modifiers
                                  , name
                                  , optInterfaceList
                                  , optConstantList
                                  );
        createTree( cTree );
    }
    RBRACE
;

//=====  Interface declaration  ================================================

/**
 * Interface declaration.
 */
interfaceDeclaration[ Token startToken
                    , List<Annotation> optAnnotationList
                    , List<Token> optModifierList
                    ]
    returns [InterfaceDeclaration interfaceDeclaration = null]
{
    Identifier name = null;
    TypeParameters optTypeParameters = null;
    List<QualifiedParameterizedIdentifier> optInterfaceList = null;
    TypeBody typeBody = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    INTERFACE
    name = identifier
    { cTree = createTree; createTree( true ); }
    optTypeParameters = optTypeParameters
    {
    	createTree( true );
    }
    (  EXTENDS optInterfaceList = optInterfaceList
     | {
            optInterfaceList = listFactory.<QualifiedParameterizedIdentifier>create();
       }
    )
    {
    	createTree( cTree );
    }
    // This is the body of an interface.
    // It can contain interface fields and extra semicolons.
    typeBodyStartToken:LBRACE
    typeBody = typeBody[typeBodyStartToken]
    {
		Token endToken = LT(1);
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        interfaceDeclaration = creator.createInterfaceDeclaration
                                       ( getPosition( startToken )
                                       , getPosition( endToken )
                                       , typeBody
                                       , javadoc
                                       , optAnnotationList
                                       , modifiers
                                       , name
                                       , optTypeParameters
                                       , optInterfaceList
                                       );
    }
    RBRACE
;

//=====  Annotation type declaration  ==========================================

/**
 * Annotation type declaration
 */
annotationTypeDeclaration[ Token startToken
                         , List<Annotation> optAnnotationList
                         , List<Token> optModifierList
                         ]
    returns [AnnotationTypeDeclaration annotationTypeDeclaration = null]
{
    Identifier name = null;
    MemberDeclaration member = null;
    List<MemberDeclaration> optMemberList = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    AT INTERFACE
    name = identifier
    // This is the body of an annotation.
    // It can contain annotation fields and extra semicolons
    typeBodyStartToken:LBRACE
    {   cTree = createTree;  
    	createTree( true );
    	optMemberList = listFactory.<MemberDeclaration>create();
    }
    (   member = annotationTypeMemberDeclaration
        {
            listFactory.add( optMemberList, member);
            createTree( true );
        }
      | SEMICOLON
    )*
    {
        createTree( cTree );
		Token endToken = LT(1);
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        annotationTypeDeclaration = creator.createAnnotationTypeDeclaration
                                            ( getPosition( startToken )
                                            , getPosition( endToken )
                                            , creator.createTypeBody( getPosition( typeBodyStartToken )
                                                                    , getPosition( endToken )
                                                                    , optMemberList
                                                                    )
                                            , javadoc                        
                                            , optAnnotationList
                                            , modifiers
                                            , name
                                            );
    }
    RBRACE
;

//=====  annotation type members  ==================================================

/**
 * Extention point to define extra categories of annotation type member declarations.
 */
annotationTypeMemberDeclaration returns [MemberDeclaration member = null]
{
    List<Annotation> optAnnotationList = null;
    List<Token> optModifierList = null;
    Type type = null;
    List<VariableDeclarator> optConstantList = null;
    Token startToken = null;
    String javadoc = null;
}
:
    {
        optAnnotationList = listFactory.<Annotation>create();
        optModifierList = listFactory.<Token>create();
        startToken = LT(1);
        javadoc = javadocFilter.getComment( startToken );
    }
    modifiers[optAnnotationList,optModifierList]
    (   member = typeDeclaration[startToken,optAnnotationList,optModifierList]
      | type = type[false]
        (   // annotation type element
            member = annotationTypeElementDeclaration
                     [ startToken
                     , optAnnotationList
                     , optModifierList
                     , type
                     ]
          | // constant declaration
              optConstantList = variableDeclarators
              {
					  Token endToken = LT(1);
                      EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
                      TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

                      member = creator.createFieldDeclaration
                                       ( getPosition( startToken )
                                       , getPosition( endToken )
                                       , javadoc
                                       , optAnnotationList
                                       , modifiers
                                       , type
                                       , optConstantList
                                       );
              }
              SEMICOLON
          )
      )
;

/**
 * Annotation type element.
 */
annotationTypeElementDeclaration[ Token startToken
                                , List<Annotation> optAnnotationList
                                , List<Token> optModifierList
                                , Type type
                                ]
returns [AnnotationTypeElementDeclaration member = null]
{
    Identifier name = null;
    ElementValue value = null;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    name = identifier
    LPARENT
    RPARENT
    ( DEFAULT
      value = elementValue
    )?
    {
		Token endToken = LT(1);
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        member = creator.createAnnotationTypeElementDeclaration
                         ( getPosition( startToken )
                         , getPosition( endToken )
                         , javadoc
                         , optAnnotationList
                         , modifiers
                         , type
                         , name
                         , value
                         );
    }
    SEMICOLON
;

//=====  Annotations  ================================================

/**
 * Zero or more annotations.
 */
optAnnotationList returns [List<Annotation> optAnnotationList = null]
{ Annotation annotation = null; boolean cTree = createTree; }
:
    {
        createTree( true );
        optAnnotationList = listFactory.<Annotation>create();
    }
    ( annotation = annotation
        {
            listFactory.add( optAnnotationList, annotation );
            createTree( true );
        }
    )*
    {
        createTree( cTree );
    }
;

annotation returns [Annotation annotation = null]
{
    QualifiedIdentifier typeName = null;
    List<ElementValuePair> optElementValuePairList = null;
    ElementValue value = null;
    Position atPosition = null;
}
:
    at:AT
    { atPosition = getPosition( at ); }
    typeName = qualifiedIdentifier
    (   LPARENT
        (   optElementValuePairList = elementValuePairs
            { Token naEndToken = LT(1); annotation = creator.createNormalAnnotation( atPosition, typeName, optElementValuePairList, getPosition( naEndToken ) ); }
          | value = elementValue 
            { Token seaEndToken = LT(1); annotation = creator.createSingleElementAnnotation( atPosition, typeName, value, getPosition( seaEndToken ) ); }
          | 
            {
              Token enaEndToken = LT(1);
              optElementValuePairList = listFactory.<ElementValuePair>create();
              annotation = creator.createNormalAnnotation( atPosition, typeName, optElementValuePairList, getPosition( enaEndToken ) );
            }
        )
        RPARENT
      | {
          annotation = creator.createMarkerAnnotation( atPosition, typeName );
        }
    )
;

elementValuePairs returns [List<ElementValuePair> optElementValuePairList = null]
{ ElementValuePair elementValuePair = null; boolean cTree = createTree; }
:
{
    createTree( true );
    optElementValuePairList = listFactory.<ElementValuePair>create();
    createTree = cTree;
}
    elementValuePair = elementValuePair
    {
        listFactory.add( optElementValuePairList, elementValuePair );
        cTree = createTree;
    }
    (	{ createTree( true ); }
        COMMA
        elementValuePair = elementValuePair
        {
            listFactory.add(optElementValuePairList, elementValuePair );
        }
    )*
    {
        createTree( cTree );
    }
;

elementValuePair returns [ElementValuePair elem_value_pair = null]
{
    Identifier ident = null;
    ElementValue elem_value = null;
}
:
    ident = identifier
    ASSIGN
    elem_value = elementValue

    { elem_value_pair = creator.createElementValuePair(ident, elem_value); }
;

elementValue returns [ElementValue elem_value = null]
:
      elem_value = conditionalExpressionElementValue
    | elem_value = annotation
    | elem_value = arrayInitializerElementValue
;

conditionalExpressionElementValue returns [ConditionalExpressionElementValue elem_value = null]
{
    ConditionalExpression expr = null;
}
:
    expr = conditionalExpression

    { elem_value = creator.createConditionalExpressionElementValue(expr); }
;

// This is an initializer used to set up an annotation member array.
arrayInitializerElementValue returns [ArrayInitializerElementValue arrayInitializerElementValue = null]
{
    ElementValue elem_value = null;
    List<ElementValue> opt_element_values = null;
    boolean cTree = createTree;
}
:
{
    createTree( true );
    opt_element_values = listFactory.<ElementValue>create();
}
    lb:LBRACE
    ( elem_value = elementValue
      {
        listFactory.add(opt_element_values, elem_value);
        createTree( true );
      }
      ( // CONFLICT: does a COMMA after an initializer start a new
        // initializer or start the option ',' at end?
        // ANTLR generates proper code by matching
        // the comma as soon as possible.
        options {warnWhenFollowAmbig = false;}:
        COMMA
        elem_value = elementValue
        {
            listFactory.add(opt_element_values, elem_value);
            createTree( true );
        }
      )*
    )?
    ( COMMA )?
    {
		Token endToken = LT(1);
        arrayInitializerElementValue = creator.createArrayInitializerElementValue( getPosition( lb ), getPosition( endToken ), opt_element_values );
        createTree( cTree );
    }
    RBRACE
;

// The two things that can initialize an annotation array element are
// a conditional expression and an annotation
// (nested annotation array initialisers are not valid)
elementValueNotArrayInitializer returns [ElementValue elem_value = null]
:
      elem_value = conditionalExpressionElementValue
    | elem_value = annotation
;

//=====  Member declarations  ==================================================

/**
 * Extention point to define extra categories of member declarations.
 */
memberDeclaration returns [MemberDeclaration member = null]
{
    List<Annotation> optAnnotationList = null;
    List<Token> optModifierList = null;
    Token startToken = null;
}
:
    {
        optAnnotationList = listFactory.<Annotation>create();
        optModifierList = listFactory.<Token>create();
        startToken = LT(1);
    }
    modifiers[optAnnotationList,optModifierList]
    member = annotatedMemberDeclaration[startToken,optAnnotationList,optModifierList]
;

/**
 * Extention point to define extra categories of member declarations
 * with annotations and modifiers.
 */
annotatedMemberDeclaration[ Token startToken
                          , List<Annotation> optAnnotationList
                          , List<Token> optModifierList
                          ]
    returns [MemberDeclaration member = null]
{
    TypeParameters optTypeParameters = null;
}
:
      member = typeDeclaration[startToken,optAnnotationList,optModifierList]
    | member = nonParameterizedMemberDeclaration[startToken,optAnnotationList,optModifierList]
    | optTypeParameters = optTypeParameters
      member = parameterizedMemberDeclaration
               [ startToken
               , optAnnotationList
               , optModifierList
               , optTypeParameters
               ]
;

/**
 * Extention point to defined extra categories of non-parameterized member declaration
 * with annotations and modifiers.
 */
nonParameterizedMemberDeclaration[ Token startToken
                                 , List<Annotation> optAnnotationList
                                 , List<Token> optModifierList
                                 ]
    returns [MemberDeclaration member = null]
:
    member = initializerDeclaration[startToken,optAnnotationList,optModifierList]
;

/**
 * Extention point to define extra categories of member declarations
 * with annotations, modifiers and type parameters.
 */
parameterizedMemberDeclaration[ Token startToken
                              , List<Annotation> optAnnotationList
                              , List<Token> optModifierList
                              , TypeParameters optTypeParameters
                              ]
    returns [MemberDeclaration member = null]
{
    OptionalType resultType = null;
}
:
    (   // constructor declaration
        member = nonTypedMemberDeclaration[ startToken
                                          , optAnnotationList
                                          , optModifierList
                                          , optTypeParameters
                                          ]
      | resultType = optionalType[false]
        member = typedMemberDeclaration[ startToken
                                       , optAnnotationList
                                       , optModifierList
                                       , optTypeParameters
                                       , resultType
                                       ]
 )
;

/**
 * Extention point to define extra categories of non-typed member declarations
 * with annotations, modifiers and type parameters.
 */
nonTypedMemberDeclaration[ Token startToken
                         , List<Annotation> optAnnotationList
                         , List<Token> optModifierList
                         , TypeParameters optTypeParameters
                         ]
    returns [MemberDeclaration member = null]
:
    member = constructorDeclaration[ startToken
                                   , optAnnotationList
                                   , optModifierList
                                   , optTypeParameters
                                   ]
;

/**
 * Extention point to define extra categories of member declarations
 * with annotations, modifiers, type parameters and type.
 */
typedMemberDeclaration[ Token startToken
                      , List<Annotation> optAnnotationList
                      , List<Token> optModifierList
                      , TypeParameters optTypeParameters
                      , OptionalType type
                      ]
    returns [MemberDeclaration member = null]
:
    member = javaTypedMemberDeclaration[ startToken
                                       , optAnnotationList
                                       , optModifierList
                                       , optTypeParameters
                                       , type
                                       ]
;

/**
 * Java member declarations which don't allow annotations and modifiers.
 */
initializerDeclaration[ Token startToken
                      , List<Annotation> optAnnotationList
                      , List<Token> optModifierList
                      ]
    returns [MemberDeclaration member = null]
{
    Block block = null;
}
:
    block = block
    {
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList, Modifier.STATIC );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );
        // block can be null after recovering from errors
        Position endPosition = block != null ? block.getEndPosition() : null;
        if( modifiers.contains( Modifier.STATIC ) )
        {
            member = creator.createStaticInitializer( getPosition( startToken ), endPosition, block );
        } else {
            member = creator.createInstanceInitializer( getPosition( startToken ), endPosition, block );
        }
    }
;

/**
 * Java member declarations which allow annotations and modifiers.
 */
javaTypedMemberDeclaration[ Token startToken
                          , List<Annotation> optAnnotationList
                          , List<Token> optModifierList
                          , TypeParameters optTypeParameters
                          , OptionalType type
                          ]
    returns [MemberDeclaration member = null]
{
    List<VariableDeclarator> variableDeclaratorList = null;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    (
      // method declaration
      member = methodDeclaration[ startToken
                                , optAnnotationList
                                , optModifierList
                                , optTypeParameters
                                , type
                                ]

    | // variable declaration
      variableDeclaratorList = variableDeclarators
      {
      	  Token endToken = LT(1);
          if( optTypeParameters != null )
          {
               //TODO: Variable declaration cannot have type parameters
               //      before variable type
          }
          if( type instanceof VoidType )
          {
              //TODO: Variable should have non-void type
          }
          EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
          TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );
          member = creator.createFieldDeclaration( getPosition( startToken )
                                                 , getPosition( endToken )
                                                 , javadoc
                                                 , optAnnotationList
                                                 , modifiers
                                                 , (Type)type
                                                 , variableDeclaratorList
                                                 );
      }
      SEMICOLON
    )
;

/**
 * One or more variable declarators.
 */
variableDeclarators returns [List<VariableDeclarator> variableDeclaratorList = null]
{
    VariableDeclarator variableDeclarator = null;
    boolean cTree = createTree;
}
:
    {
    	createTree( true );
        variableDeclaratorList = listFactory.<VariableDeclarator>create();
        createTree( cTree );
    }
    variableDeclarator = variableDeclarator
    {
    	listFactory.add( variableDeclaratorList, variableDeclarator );
    	cTree = createTree;
    }
    ( { createTree( true ); }
   	  COMMA
      variableDeclarator = variableDeclarator
      {
        listFactory.add( variableDeclaratorList, variableDeclarator );
      }
    )*
    { createTree( cTree ); }
;

variableDeclaratorId returns [VariableDeclaratorId variableDeclaratorId = null ]
{
    Identifier name = null;
    int dim = 0;
}
:
    name = identifier
    ( //options{greedy=true;}: // match as many as possible
      LBRACKET  endToken:RBRACKET { dim++; }
    )*
    {
        if( createTree )
        {
            Position endPosition =   endToken == null && name != null
                                   ?   endPosition = endVisitor.getEndPosition( name ) 
                                     : getPosition( endToken );
            variableDeclaratorId = creator.createVariableDeclaratorId( endPosition, name, dim );
        }
    }
;


/**
 * Variable declarator.
 */
variableDeclarator returns [VariableDeclarator variableDeclarator = null]
{
    VariableDeclaratorId variableDeclaratorId = null;
    VariableInitializer initializer = null;
}
:
    variableDeclaratorId = variableDeclaratorId
    ( ASSIGN
      initializer = variableInitializer
    )?
    {
        variableDeclarator = creator.createVariableDeclarator( variableDeclaratorId, initializer );
    }
;

/**
 * Variable initializer.
 */
variableInitializer returns [VariableInitializer variableInitializer = null]
{
    Expression expression = null;
}
:
      expression = expression
      { variableInitializer = creator.createExpressionVariableInitializer( expression ); }
    | variableInitializer = arrayInitializer
;

/**
 * Array initializer.
 */
arrayInitializer returns [ArrayInitializer arrayInitializer = null]
{
    VariableInitializer initializer = null;
    List<VariableInitializer> optInitializerList = null;
    boolean cTree = createTree;
}
:
    {
        createTree( true );
        optInitializerList = listFactory.<VariableInitializer>create();
    }
    lb:LBRACE
    ( initializer = variableInitializer
      {
        listFactory.add( optInitializerList, initializer );
        createTree( true );
      }
      ( // CONFLICT: does a COMMA after an initializer start a new
        // initializer or start the option ',' at end?
        // ANTLR generates proper code by matching
        // the comma as soon as possible.
        options { warnWhenFollowAmbig = false; }:
        COMMA
        initializer = variableInitializer
        {
            listFactory.add( optInitializerList, initializer );
            createTree( true );
        }
      )*
    )?
    ( COMMA )?
    {
    	Token endToken = LT(1);
        arrayInitializer = creator.createArrayInitializer( getPosition( lb ), getPosition( endToken ), optInitializerList );
        createTree( cTree );
    }
    RBRACE
;

/**
 * Method header.
 */
methodHeader[ TypeParameters optTypeParameters
            , OptionalType type
            ]
    returns [MethodHeader methodHeader = null]
{
    Identifier name = null;
    List<OperationFormalParameter> optFormalParameterList = null;
    QualifiedIdentifier exceptionType = null;
    List<QualifiedIdentifier> optExceptionList = null;
    boolean arity = false;
    int dim = 0;
    Position mhEndPosition = null;
    boolean cTree = createTree;
}
:
    name = identifier
    // parse the formal parameter declarations
    LPARENT
    { 
    	cTree = createTree;
		createTree( true );
    	optFormalParameterList = listFactory.<OperationFormalParameter>create();
    	createTree( cTree );
    }
    arity = formalParameters[optFormalParameterList]
    { mhEndPosition = getPosition( LT(1) ); }
    RPARENT
    //For compatibility with older versions of the Java platform
    // a declaration form for a method that returns an array
    // is allowed to place (some or all of) the empty bracket pairs
    // that form the declaration of the array type after the parameter list.
    ( LBRACKET endToken:RBRACKET { dim++; mhEndPosition = getPosition( endToken ); } )*
    // get the list of exceptions that this method is declared to throw
    {
    	cTree = createTree;
        createTree( true );
        optExceptionList = listFactory.<QualifiedIdentifier>create();
    }
    ( THROWS
      exceptionType = qualifiedIdentifier
      {
            listFactory.add( optExceptionList, exceptionType );
            createTree( true );
      }
      ( COMMA
        exceptionType = qualifiedIdentifier
        {
            listFactory.add( optExceptionList, exceptionType );
            createTree( true );
        }
      )*
    )?
    {
    	createTree( cTree );
    	
        if( dim > 0 )
        {
            // If a method declaration contains the empty bracket pairs (see above)
            // increase dimension of the array type of the result if it is an array type
            // or change the result type to an array of corresponding type.
            if( type instanceof ArrayType )
            {
                ArrayType arrayType = (ArrayType)type;
                arrayType.setDim( arrayType.getDim() + dim );
            } else
            if( type instanceof VoidType )
            {
                // TODO: report error
            } else {
                type = creator.createArrayType( getPosition( endToken ), (Type)type, dim );
            }
        }
        if( optExceptionList != null && optExceptionList.size() > 0 )
        {
            mhEndPosition = endVisitor.getEndPosition( optExceptionList );
        }
        methodHeader = creator.createMethodHeader
                               ( mhEndPosition
                               , optTypeParameters
                               , type
                               , name
                               , optFormalParameterList
                               , arity
                               , optExceptionList
                               );
         createTree( cTree );
    }
;

/**
 * Method declaration
  */
methodDeclaration[ Token startToken
                 , List<Annotation> optAnnotationList
                 , List<Token> optModifierList
                 , TypeParameters optTypeParameters
                 , OptionalType type
                 ]
returns [MemberDeclaration methodDeclaration = null]
{
    MethodHeader methodHeader = null;
    Block body = null;
	boolean cTree = createTree;
	String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    methodHeader = methodHeader[optTypeParameters,type]
    {
    	cTree = createTree;
   	}
    (   body = block
      | endToken:SEMICOLON
    )
    {
    	createTree( cTree );    	
    	removeInappropriates( optModifierList, methodDeclarationInappropriatesModifiers );
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );
        Position endPosition = body != null ? endPosition = body.getEndPosition() : getPosition( endToken );

        methodDeclaration = creator.createMethodDeclaration
                                    ( getPosition( startToken )
                                    , endPosition
                                    , javadoc
                                    , optAnnotationList
                                    , modifiers
                                    , methodHeader
                                    , body
                                    );
    }
;

// A list of formal parameters: zero or more parameters.
// If a parameter is variable length (e.g. String... myArg)
// it is the right-most parameter
formalParameters[List<OperationFormalParameter> optFormalParameterList]
returns [boolean arity = false]
{
    OperationFormalParameter param = null;
    boolean cTree = createTree;
}
:
	{ createTree( true ); }
  ( param = formalParameter[optFormalParameterList,true]
    { if( createTree && param == null ) arity = true; }
    ( // TODO: remove semantic predicate, check arity
      // of previous parameter after next COMMA is matched
      // and report error
      { !arity }?
      COMMA param = formalParameter[optFormalParameterList,true]
      { if( createTree && param == null ) arity = true;  createTree( true ); }
    )*
  )?
	{ createTree( cTree ); }
;

// A formal parameter
//
// This rule is used in two contexts:
// 1. in parameter list (formalParameters)
// 2. in statements where single parameter is declared
//
// if !isList
//     VARARG (...) not allowed
//     and parameter is returned in 'param'
// else
//     VARARG is allowed
//     parameter is added in optFormalParameterList
//     if VARARG found, param == null
//
formalParameter[ List<OperationFormalParameter> optFormalParameterList
               , boolean isList
               ]
    returns [OperationFormalParameter param=null]
{
    List<Annotation> optAnnotationList = null;
    List<Token> optModifierList = null;
    Type type = null;
    boolean arity = false;
    Identifier name = null;
    int dim = 0;
    Token startToken = null;
    Token endToken = null;
}
:
    {
        startToken = LT(1);
        optAnnotationList = listFactory.<Annotation>create();
        optModifierList = listFactory.<Token>create();
    }
    modifiers[optAnnotationList,optModifierList]
    type = type[false]
    ( { isList }? { endToken = LT(1); } VARARG { arity = true; } )?
    name = identifier
    ( { !arity }? LBRACKET { endToken = LT(1); } RBRACKET { dim++; } )*
    {
        // If a parameter declaration contains the empty bracket pairs after
        // its name, increase dimension of the array type of a parameter
        // if it is an array type or change its type to an array
        // of the corresponding type.
        if( dim > 0 )
        {
            if( type instanceof ArrayType )
            {
                ArrayType t = (ArrayType)type;
                t.setDim( t.getDim() + dim );
            } else {
                type = creator.createArrayType( getPosition( endToken ), type, dim );
            }
        }
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        if( arity )
        {
            param = creator.createVariableArityFormalParameter( getPosition( startToken )
                                                              , optAnnotationList
                                                              , modifiers
                                                              , type
                                                              , name
                                                              );
        } else {
            param = creator.createFormalParameter( getPosition( startToken )
                                                 , optAnnotationList
                                                 , modifiers
                                                 , type
                                                 , name
                                                 );
        }
        if( isList )
        {
            listFactory.add( optFormalParameterList, param );
        }
        if( arity ) param = null;

    }
;

/*
parameterOrLocalVarModifier[List/Annotation/ optAnnotationList]
returns [boolean is_final = false]
{
    Annotation annotation = null;
    boolean cTree = createTree;
}
:
    ( //"final" can appear amongst annotations in any order -
      // greedily consume any preceding annotations
      // to shut non-determinism warnings off
      options { greedy = true; }:
		{ createTree( true ); }
      annotation = annotation
      {
        listFactory.add( optAnnotationList, annotation );
      }
    )*
    { createTree( true ); }
    ( FINAL
      { is_final = true; }
    )?
    ( 
	    { createTree( true ); }
    	annotation = annotation
      {
        listFactory.add( optAnnotationList, annotation );
      }
    )*
    {
        createTree( cTree );
    }
;*/

/**
 * Constructor declaration.
 */
constructorDeclaration[ Token startToken
                      , List<Annotation> optAnnotationList
                      , List<Token> optModifierList
                      , TypeParameters optTypeParameters
                      ]
returns [ConstructorDeclaration constructorDeclaration = null]
{
    Identifier name = null;
    boolean arity = false;
    List<OperationFormalParameter> optFormalParameterList = null;
    QualifiedIdentifier exceptionType = null;
    List<QualifiedIdentifier> optExceptionList = null;
    ExplicitConstructorInvocation optExplicitConstructorCall = null;
    BlockStatement statement = null;
    List<BlockStatement> optBlockStatementList = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    name = identifier

    // parse the formal parameter declarations
    LPARENT
    {
    	cTree = createTree;
	    createTree( true );
        optFormalParameterList = listFactory.<OperationFormalParameter>create();
    }
    arity = formalParameters[optFormalParameterList]
    RPARENT

    // get the list of exceptions that this method is declared to throw
    { 
    	cTree = createTree;
    	createTree( true );
    	optExceptionList = listFactory.<QualifiedIdentifier>create();
    }
    ( THROWS
      exceptionType = qualifiedIdentifier
      {
        listFactory.add( optExceptionList, exceptionType );
        createTree( true );
      }
      ( COMMA
        exceptionType = qualifiedIdentifier
        {
            listFactory.add( optExceptionList, exceptionType );
            createTree( true );
        }
      )*
    )?
    LBRACE
    optExplicitConstructorCall = explicitConstructorInvocation
    {
    	cTree = createTree;
    	createTree( true ); 
    	optBlockStatementList = listFactory.<BlockStatement>create();
    }
    ( statement = blockStatement
      {
        listFactory.add( optBlockStatementList, statement );
        createTree( true );
      }
    )*
    {
        createTree( cTree );    	
    	Token endToken = LT(1);
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        constructorDeclaration = creator.createConstructorDeclaration
                                         ( getPosition( startToken )
                                         , getPosition( endToken )
                                         , javadoc
                                         , optAnnotationList
                                         , modifiers
                                         , optTypeParameters
                                         , name
                                         , optFormalParameterList
                                         , arity
                                         , optExceptionList
                                         , optExplicitConstructorCall
                                         , optBlockStatementList
                                         );
    }
    RBRACE
;

// Catch constructor calls
explicitConstructorInvocation returns [ExplicitConstructorInvocation expl_constr_invoc = null]
{
    TypeArguments optTypeArguments = null;
    List<Expression> opt_args = null;
    Primary primary_expr = null;
    Token startToken = null;
}
:
    { startToken = LT(1); }
    (
      // Resolving nondeterminism between "THIS();"
      // and primary in the last alternative of this rule
      ( optTypeArguments[false, false] ( THIS | SUPER ) LPARENT )=>
      ( optTypeArguments = optTypeArguments[false, false]
        // "this(...)" constructor call
        (  THIS
           LPARENT
           opt_args = optExpressionList
           RPARENT
           tEndToken:SEMICOLON
           {
               expl_constr_invoc
                   = creator.createThisExplicitConstructorInvocation
                             ( getPosition( startToken ), getPosition( tEndToken ), optTypeArguments, opt_args );
           }
         | // "super(...)" constructor call
           SUPER
           LPARENT
           opt_args = optExpressionList
           RPARENT
           sEndToken:SEMICOLON
           {
               expl_constr_invoc = creator.createSuperExplicitConstructorInvocation
                                           ( getPosition( startToken )
                                           , getPosition( sEndToken )
                                           , optTypeArguments
                                           , opt_args
                                           , null
                                           );
           }
        )
      )
    | // "expr.super(...)" constructor call
      // Resolving nondeteminism between "expr.super(...)"
      // and primaries begining with "SUPER"
      ( primary DOT optTypeArguments[false, false] SUPER LPARENT )=>
      primary_expr = primary
      DOT
      optTypeArguments = optTypeArguments[false, false]
      SUPER
      LPARENT
      opt_args = optExpressionList
      RPARENT
      secEndToken:SEMICOLON
      {
          expl_constr_invoc = creator.createSuperExplicitConstructorInvocation
                                      ( getPosition( startToken )
                                      , getPosition( secEndToken )
                                      , optTypeArguments
                                      , opt_args
                                      , primary_expr
                                      );
      }
    | //nothing
    )
;

//=====  Enum member declarations  =====================================================

/**
 * Enum constant.
 */
enumConstant returns [EnumConstant enumConstant = null]
{
    List<Annotation> optAnnotationList = null;
    Identifier name = null;
    List<Expression> optArgumentList = null;
    TypeBody optTypeBody = null;
    Token startToken = null;
    Token endToken = null;
    boolean cTree = createTree;
}
:
    { startToken = LT(1); }
    optAnnotationList = optAnnotationList
    { endToken = LT(1); }
    name = identifier
    { cTree = createTree; createTree( true ); }
    (   LPARENT
        optArgumentList = optExpressionList
        { endToken = LT(1); }
        RPARENT
      | { optArgumentList = listFactory.<Expression>create(); }
    )
    (
    	{ createTree( true ); }
      // The class-like body of an enum constant
      typeBodyStartToken:LBRACE
      optTypeBody = typeBody[typeBodyStartToken]
      { endToken = LT(1); }
      RBRACE
    )?
    {
    	createTree( cTree );
        enumConstant = creator.createEnumConstant( getPosition( startToken )
                                                 , getPosition( endToken )
                                                 , optAnnotationList
                                                 , name
                                                 , optArgumentList
                                                 , optTypeBody
                                                 );
    }
;

//=====  Statements  ================================================

// Compound statement. This is used in many contexts:
// Inside a class definition prefixed with "static":
// it is a class initializer
// Inside a class definition without "static":
// it is an instance initializer
// As the body of a method
// As a completely indepdent braced block of code inside a method
// it starts a new scope for variable definitions
block returns [Block block = null]
{
    BlockStatement statement = null;
    List<BlockStatement> optStatementList = null;
    boolean cTree = createTree;
    Token endToken = null;
}
:
    {   createTree( true );
        optStatementList = listFactory.<BlockStatement>create(); }
    startToken:LBRACE
    ( statement = blockStatement
      { listFactory.add( optStatementList, statement ); createTree( true ); }
    )*
    { endToken = LT(1);
      block = creator.createBlock( getPosition( startToken ), getPosition( endToken ), optStatementList );
      createTree( cTree );
    }
    RBRACE
;

// Any statement, local variable declaration,
// or local class or enum declaration
blockStatement returns [BlockStatement statement = null]
{
    LocalVariableDeclaration localVariableDeclaration = null;
    List<Annotation> optAnnotationList = null;
    List<Token> optModifierList = null;
    TypeDeclaration typeDeclaration = null;
    Token startToken = null;
}
:
      ( AT | modifier | CLASS | ENUM | type[false] identifier )=>
      // Local class, enum or variable declaration
      // synchronized statement is not matched by this alternative
      // because of '(' is not in second-symbol-set
      (
        {
            optAnnotationList = listFactory.<Annotation>create();
            optModifierList = listFactory.<Token>create();
            startToken = LT(1);
        }
        modifiers[optAnnotationList,optModifierList]
        (
          // Local variable declaration
           // ( type identifier )=>
           localVariableDeclaration = localVariableDeclaration[startToken,optAnnotationList,optModifierList]
           SEMICOLON
           { statement = creator.createLocalVariableDeclarationStatement( localVariableDeclaration ); }
         | // Local class declaration
           typeDeclaration = classDeclaration[startToken,optAnnotationList,optModifierList]
           { statement = creator.createClassDeclarationStatement( (ClassDeclaration)typeDeclaration ); }
         | // Local enum declaration
           typeDeclaration = enumDeclaration[startToken,optAnnotationList,optModifierList]
           { statement = creator.createEnumDeclarationStatement( (EnumDeclaration)typeDeclaration ); }
        )
      )
    | // Ordinary statements
      statement = statement
;

// Local variable declaration looks like as formal parameter one
localVariableDeclaration[ Token startToken
                        , List<Annotation> optAnnotationList
                        , List<Token> optModifierList
                        ]
returns [LocalVariableDeclaration local_var_decl = null]
{
    Type t = null;
    List<VariableDeclarator> var_decls = null;
}
:
    t = type[false]
    var_decls = variableDeclarators
    {
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );
        local_var_decl = creator.createLocalVariableDeclaration
                                 ( getPosition( startToken ), optAnnotationList, modifiers, t, var_decls );
    }
;

// Any statement
statement returns [Statement stmnt = null]
:
    stmnt = javaStatement
;

javaStatement returns [Statement stmnt = null]
{
    Identifier label = null;
    Statement inner_statement = null;
    Statement else_statement = null;
    Expression expr = null;
    Expression message_expr = null;
    List<SwitchElement> switch_elements = null;
}
:
      // A list of statements in curly braces -- start a new scope
      stmnt = block
    | // An empty statement
      t:SEMICOLON
      { stmnt = creator.createEmptyStatement( getPosition( t ) ); }
    | // A labeled statement
      label = identifier COLON inner_statement = statement
      { stmnt = creator.createLabeledStatement(label, inner_statement); }
    | // An expression statement. This could be a method call,
      // assignment statement, or any other expression evaluated for
      // side-effects.
      expr = statementExpression endToken:SEMICOLON
      { stmnt = creator.createExpressionStatement( getPosition( endToken ), expr ); }
    | // if-else statement
      ift:IF
      LPARENT expr = expression RPARENT
      inner_statement = statement
      ( // CONFLICT: the old "dangling-else" problem...
        // ANTLR generates proper code matching
        // as soon as possible. Hush warning.
        options { warnWhenFollowAmbig = false; }:
        ELSE
        else_statement = statement
      )?
      {
          stmnt = creator.createIfStatement( getPosition( ift )
                                           , expr
                                           , inner_statement
                                           , else_statement
                                           );
      }
    | // Asserts (for 1.4 compatibility)
      assertStartToken:ASSERT
      expr = expression
      ( COLON message_expr = expression )?
      assertEndToken:SEMICOLON
      { stmnt = creator.createAssertStatement( getPosition( assertStartToken ), getPosition( assertEndToken ), expr, message_expr ); }
    | // switch/case statement
      switchStartToken:SWITCH
      LPARENT expr = expression RPARENT
      LBRACE
      ( switch_elements = switchElements | { switch_elements = listFactory.create(); } )
      switchEndToken:RBRACE
      { stmnt = creator.createSwitchStatement( getPosition( switchStartToken ), getPosition( switchEndToken ), expr, switch_elements ); }
    | // while statement
      wht:WHILE
      LPARENT expr = expression RPARENT
      inner_statement = statement
      { stmnt = creator.createWhileStatement( getPosition( wht ), expr, inner_statement ); }
    | // do-while statement
      doStartToken:DO inner_statement = statement
      WHILE LPARENT expr = expression RPARENT doEndToken:SEMICOLON
      { stmnt = creator.createDoStatement( getPosition( doStartToken ), getPosition( doEndToken ), inner_statement, expr ); }
    | // for statement
      frt:FOR
      (  ( LPARENT formalParameter[null,true] COLON )=>
         stmnt = enhancedForStatement[frt]
       | stmnt = basicForStatement[frt]
      )
    | // get out of a loop (or switch)
      brt:BREAK ( label = identifier )? breakEndToken:SEMICOLON
      { stmnt = creator.createBreakStatement( getPosition( brt ), getPosition( breakEndToken ), label ); }
    | // do next iteration of a loop
      cnt:CONTINUE ( label = identifier )? continueEndToken:SEMICOLON
      { stmnt = creator.createContinueStatement( getPosition( cnt ), getPosition( continueEndToken ), label ); }
    | // Return an expression
      stmnt = returnStatement    
    | // Throw an exception
      tht:THROW expr = expression throwEndToken:SEMICOLON
      { stmnt = creator.createThrowStatement( getPosition( tht ), getPosition( throwEndToken ), expr ); }
    | // Synchronize a statement
      syt:SYNCHRONIZED LPARENT expr = expression RPARENT
      stmnt = block
      { stmnt = creator.createSynchronizedStatement( getPosition( syt ), expr, (Block)stmnt ); }
    | // An exception try-catch block
      stmnt = tryStatement
;

returnStatement returns [Statement stmnt = null]
{
    Expression expr = null;
}
:
    // Return an expression
    rtt:RETURN ( expr = expression )? returnEndToken:SEMICOLON
    { stmnt = creator.createReturnStatement( getPosition( rtt ), getPosition( returnEndToken ), expr ); }
;

// A statement expression. This could be a method call,
// assignment statement, or any other expression evaluated for
// side-effects.
statementExpression returns [Expression expr = null]
:
    expr = expression
;

// Switch elements are list block statements or one kind of label
// ("case <ENUM_CONSTANT_IDENT>:", "case <CONSTANT_EXPR>:" or "default:").
// Before each list of block statements should be one or more labels.
switchElements returns [List<SwitchElement> switch_elements = null]
{
    BlockStatement statement = null;
    List<BlockStatement> statements = null;
    Expression const_expr = null;
    boolean cTree = createTree;
}
:
{
    createTree( true );
    switch_elements = listFactory.<SwitchElement>create();
    statements = listFactory.<BlockStatement>create();
}
    ( ( // CONFLICT: to which case group do the statements bind?
        // ANTLR generates proper code: it groups the
        // many "case"/"default" labels together then
        // follows them with the statements
        options { greedy = true; }:
        (  cst:CASE
           // Constant expression or enum constant reference label
           const_expr = expression
           clt1:COLON
           {
           	listFactory.add( switch_elements
           	               , creator.createSwitchLabel( getPosition( cst )
           	                                          , getPosition( clt1 )
           	                                          , const_expr 
           	                                          ) 
           	               );
           	createTree( true );
           }
         | // Default label
           dft:DEFAULT
           clt2:COLON
           {
           	listFactory.add( switch_elements, creator.createDefaultSwitchLabel( getPosition( dft ), getPosition( clt2 ) ) );
           	createTree( true );
           }
        )
      )+
      ( statement = blockStatement
        {
        	listFactory.add(statements, statement);
        	createTree( true );
        }
      )*
      {
          if( statements != null && statements.size() != 0 )
          {
              listFactory.add(switch_elements, creator.createStatementSwitchElement(statements));
              statements = listFactory.create();
          }
          createTree( true );
      }
    )+
    {
        createTree( cTree );
    }
;

// Traditional for-loop ("for(...;...;...)")
basicForStatement[Token startToken] returns [BasicForStatement for_loop = null]
{
    ForInit for_init = null;
    Expression cond_expr = null;
    List<Expression> update_exprs = null;
    Statement for_body = null;
}
:
    LPARENT
    ( for_init = forInit )?
    SEMICOLON
    ( cond_expr = expression )?
    SEMICOLON
    update_exprs = optExpressionList
    RPARENT
    for_body = statement

    { for_loop = creator.createBasicForStatement(getPosition( startToken ), for_body, for_init, cond_expr, update_exprs); }
;

// The initializer for a for-loop
forInit returns [ForInit for_init = null]
{
    List<Annotation> optAnnotationList = null;
    List<Token> optModifierList = null;
    LocalVariableDeclaration var_decl = null;
    List<Expression> exprs = null;
    Token startToken = null;
}
:
      // if it looks like a local variable declaration, it is
      ( modifiers[null,null] localVariableDeclaration[null,null,null] )=>
      {
          optAnnotationList = listFactory.<Annotation>create();
          optModifierList = listFactory.<Token>create();
          startToken = LT(1);
      }
      modifiers[optAnnotationList,optModifierList]
      var_decl = localVariableDeclaration[startToken,optAnnotationList,optModifierList]
      { for_init = creator.createLocalVariableDeclarationForInit(var_decl); }
    | // otherwise it should be an expression list
      exprs = expressionList
      { 
        if( exprs != null && exprs.size() == 0 )
        {
          listFactory.add( exprs, creator.createNullLiteral( getPosition(LT(0)), "null" ) );
        }
        for_init = creator.createStatementExpressionListForInit(exprs);
      }
;

// For-loop with "foreach clause"
enhancedForStatement[Token startToken] returns[EnhancedForStatement foreach_loop = null]
{
    OperationFormalParameter param = null;
    Expression expr = null;
    Statement for_body = null;
}
:
    LPARENT param = formalParameter[null,false] COLON expr = expression RPARENT
    for_body = statement

    { foreach_loop = creator.createEnhancedForStatement
                             ( getPosition( startToken )
                             , for_body
                             , param
                             , expr
                             );
    }
;

// Try statement contains try block and exactly one catch clause
// (consisting of a formal parameter declaration and catch block)
// or try and finally blocks and optional list of catch clauses.
tryStatement returns [TryStatement try_statement = null]
{
    Block try_block = null;
    OperationFormalParameter param = null;
    Block catch_block = null;
    List<CatchClause> catch_clauses = null;
    Block finally_block = null;
    boolean cTree = createTree;
}
:
{
	createTree( true );
    catch_clauses = listFactory.<CatchClause>create();
    createTree( cTree );
}
      trt:TRY try_block = block
      { cTree = createTree; }
      ( { createTree( true ); }
      	cat:CATCH LPARENT param = formalParameter[null,false] RPARENT catch_block = block
        {
	        listFactory.add( catch_clauses, creator.createCatchClause( getPosition( cat ), param, catch_block ) );
	        createTree( true );
       	}
      )*
      ( FINALLY finally_block = block )?

      {
          if(    createTree
              && catch_clauses.size() == 0
              && finally_block == null
            )
          {
              Position position = endVisitor.getEndPosition( try_block );
              mbox.error( file
                        , position
                        , new Message.TryNoCatchOrFinally( getPosition( trt ) )
                        );
              stopCreateTree();
          }
          else
          {
	          try_statement = creator.createTryStatement( getPosition( trt ), try_block, catch_clauses, finally_block );
	          createTree( cTree );
	      }
      }
;

//=====  Expressions  ================================================

// Note that most of the following expressions follow the pattern
//   thisLevelExpression :
//         nextHigherPrecedenceExpression
//                 (OPERATOR nextHigherPrecedenceExpression)*
// which is a standard recursive definition for a parsing an expression.
// The operators in java have the following precedences:
//      lowest  (13)  = *= /= %= += -= <<= >>= >>>= &= ^= |=
//                      (12)  ?:
//                      (11)  ||
//                      (10)  &&
//                      ( 9)  |
//                      ( 8)  ^
//                      ( 7)  &
//                      ( 6)  == !=
//                      ( 5)  < <= > >=
//                      ( 4)  << >>
//                      ( 3)  +(binary) -(binary)
//                      ( 2)  * / %
//                      ( 1)  ++ -- +(unary) -(unary)  ~  !  (type)
//                                []   () (method call)  . (dot -- identifier qualification)
//                                new   ()  (explicit parenthesis)
//
// the last two are not usually on a precedence chart. They are put in
// to point out that new has a higher precedence than '.', so it can be
// validy used
//       new Frame().show()
//
// Note that the above precedence levels map to the rules below.
// Once a precedence chart is haven, writing the appropriate rules as below
// is usually very straightfoward

// The root of all expressions (assignment (level 13))
expression returns [Expression expr = null]
:
    expr = assignment
;

// An assignment (level 13))
assignment returns [Expression expr = null]
{
    ConditionalExpression cond_expr = null;
    AssignmentOperator op = null;
    Expression right_expr = null;
    Token signToken = null;
}
:
    cond_expr = conditionalExpression
    ( { signToken = LT(1); } op = assignmentOperator right_expr = expression )?
    {
        if( right_expr == null )
        {
            expr = cond_expr;
        } else {
            expr = creator.createAssignment( cond_expr, getPosition( signToken ), op, right_expr );
        }
    }
;

// This is a list of expressions.
expressionList returns [List<Expression> expressionList = null]
{
    Expression expr = null;
    boolean cTree = createTree;
}
:
    { createTree( true ); expressionList = listFactory.<Expression>create(); }
            expr = expression { listFactory.add( expressionList, expr ); createTree( true ); }
    ( COMMA expr = expression { listFactory.add( expressionList, expr ); createTree( true ); } )*
    { createTree( cTree ); }
;

optExpressionList returns [List<Expression> optExpressionList = null]
:
    optExpressionList = expressionList | { optExpressionList = listFactory.create(); }
;

assignmentOperator returns [AssignmentOperator op = null]
:
      ASSIGN         { op = AssignmentOperator.ASSIGN ; }
    | PLUS_ASSIGN    { op = AssignmentOperator.ASSIGN_PLUS ; }
    | MINUS_ASSIGN   { op = AssignmentOperator.ASSIGN_MINUS ; }
    | MULT_ASSIGN    { op = AssignmentOperator.ASSIGN_MULTIPLY ; }
    | DIVIDE_ASSIGN  { op = AssignmentOperator.ASSIGN_DIVIDE ; }
    | REM_ASSIGN     { op = AssignmentOperator.ASSIGN_REMAINDER ; }
    | AND_ASSIGN     { op = AssignmentOperator.ASSIGN_AND ; }
    | OR_ASSIGN      { op = AssignmentOperator.ASSIGN_OR ; }
    | XOR_ASSIGN     { op = AssignmentOperator.ASSIGN_XOR ; }
    | LSHIFT_ASSIGN  { op = AssignmentOperator.ASSIGN_LSHIFT ; }
    | RSHIFT_ASSIGN  { op = AssignmentOperator.ASSIGN_RSHIFT ; }
    | URSHIFT_ASSIGN { op = AssignmentOperator.ASSIGN_URSHIFT ; }
;

// conditional test (level 12)
conditionalExpression returns [ConditionalExpression expr = null]
{
    ConditionalOrExpression condition_expr = null;
    ConditionalExpression else_expr = null;
    Expression then_expr = null;
}
:
    condition_expr = conditionalOrExpression
    ( q:QUESTION then_expr = expression COLON else_expr = conditionalExpression )?
    {
        if( then_expr == null )
        {
            expr = condition_expr;
        } else {
            expr = creator.createStrictConditionalExpression
                           ( condition_expr, getPosition( q ), then_expr, else_expr );
        }
    }
;

// logical or (||) (level 11)
conditionalOrExpression returns [ConditionalOrExpression expr = null]
{
    ConditionalAndExpression right = null;
//    boolean cTree = createTree;
}
:
    expr = conditionalAndExpression
//    {
//    	cTree = createTree;
//    }
    ( //{ createTree( true ); }
    	signToken:SC_OR right = conditionalAndExpression
      {
          expr = creator.createStrictConditionalOrExpression( expr, getPosition( signToken ), right );
      }
    )*
//    {
//    	createTree( cTree );
//    }
;

// logical and (&&) (level 10)
conditionalAndExpression returns [ConditionalAndExpression expr = null]
{
    InclusiveOrExpression right = null;
}
:
    expr = inclusiveOrExpression
    ( signToken:SC_AND right = inclusiveOrExpression
      {
          expr = creator.createStrictConditionalAndExpression( expr, getPosition( signToken ), right );
      }
    )*
;

// bitwise or non-short-circuiting or (|) (level 9)
inclusiveOrExpression returns [InclusiveOrExpression expr = null]
{
    ExclusiveOrExpression right = null;
}
:
    expr = exclusiveOrExpression
    ( signToken:BIT_OR right = exclusiveOrExpression
      {
          expr = creator.createStrictInclusiveOrExpression( expr, getPosition( signToken ), right );
      }
    )*
;

// exclusive or (^) (level 8)
exclusiveOrExpression returns [ExclusiveOrExpression expr = null]
{
    AndExpression right = null;
}
:
    expr = andExpression
    ( signToken:XOR right = andExpression
      {
          expr = creator.createStrictExclusiveOrExpression( expr, getPosition( signToken ), right );
      }
    )*
;

// bitwise or non-short-circuiting and (&) (level 7)
andExpression returns [AndExpression expr = null]
{
    EqualityExpression right = null;
}
:
    expr = equalityExpression
    ( signToken:BIT_AND right = equalityExpression
      {
          expr = creator.createStrictAndExpression( expr, getPosition( signToken ), right );
      }
    )*
;

// equality/inequality (==/!=) (level 6)
equalityExpression returns [EqualityExpression expr = null]
{
    EqualitySign op = null;
    RelationalExpression right = null;
    Token signToken = null;
}
:
    expr = relationalExpression
    ( { signToken = LT(1); }
      (  EQ { op = EqualitySign.EQ; }
       | NE { op = EqualitySign.NE; }
      )
      right = relationalExpression
      {
          expr = creator.createStrictEqualityExpression( expr, getPosition( signToken ), op, right );
      }
    )*
;

// boolean relational expressions (level 5)
relationalExpression returns [RelationalExpression expr = null]
{
    RelationalSign op = null;
    ShiftExpression right = null;
    Type sourceType = null;
    Token signToken = null;
}
:
    expr = shiftExpression
    (   ( { signToken = LT(1); }
          (   LTHN { op = RelationalSign.LT; }
            | GTHN { op = RelationalSign.GT; }
            | LE   { op = RelationalSign.LE; }
            | GE   { op = RelationalSign.GE; }
          )
          right = shiftExpression
          {
              expr = creator.createStrictRelationalExpression( expr, getPosition( signToken ), op, right );
          }
        )*
      | ( INSTANCEOF sourceType = type[false]
          {
              if( sourceType != null )
         	  {
	         	 if ( sourceType instanceof ReferenceType )
		         {
		              expr = creator.createInstanceOfExpression( expr, (ReferenceType) sourceType ); 
		         }
		         else
		         {
		         	mbox.error( file, startVisitor.getStartPosition( sourceType ), new Message.ReferenceTypeExpectedButNotPrimitive() );
		         	stopCreateTree();
		         }
		      }
          }
        )+
    )
;

// bit shift expressions (level 4)
shiftExpression returns [ShiftExpression expr = null]
{
    ShiftSign op = null;
    AdditiveExpression right = null;
    Token signToken = null;
}
:
    expr = additiveExpression
    ( { signToken = LT(1); }
      (  LSHIFT  { op = ShiftSign.LEFT; }
       | RSHIFT  { op = ShiftSign.RIGHT; }
       | URSHIFT { op = ShiftSign.URIGHT; }
      )
      right = additiveExpression
      {
          expr = creator.createStrictShiftExpression( expr, getPosition( signToken ), op, right );
      }
    )*
;

// binary addition/subtraction (level 3)
additiveExpression returns [AdditiveExpression expr = null]
{
    AdditiveSign op = null;
    MultiplicativeExpression right = null;
    Token signToken = null;
}
:
    expr = multiplicativeExpression
    ( { signToken = LT(1); }
      (  PLUS  { op = AdditiveSign.PLUS; }
       | MINUS { op = AdditiveSign.MINUS; }
      )
      right = multiplicativeExpression
      {
          expr = creator.createStrictAdditiveExpression( expr, getPosition( signToken ), op, right );
      }
    )*
;

// multiplication/division/modulo (level 2)
multiplicativeExpression returns [MultiplicativeExpression expr = null]
{
    MultiplicativeSign op = null;
    UnaryExpression right = null;
    Token signToken = null;
}
:
    expr = unaryExpression
    ( { signToken = LT(1); }
      (  MULT   { op = MultiplicativeSign.MULT; }
       | DIVIDE { op = MultiplicativeSign.DIV; }
       | REM    { op = MultiplicativeSign.REMAINDER; }
      )
      right = unaryExpression
      {
          expr = creator.createStrictMultiplicativeExpression( expr, getPosition( signToken ), op, right );
      }
    )*
;

unaryExpression returns [UnaryExpression expr = null]
{
    UnaryExpression e = null;
    Token startToken = null;
}
:
    { startToken = LT(1); }
    (   INCR e = unaryExpression
        { expr = creator.createPreAdditiveExpression( getPosition( startToken ), AdditiveSign.PLUS, e ); }
      | DECR e = unaryExpression
        { expr = creator.createPreAdditiveExpression( getPosition( startToken ), AdditiveSign.MINUS, e ); }
      | PLUS e = unaryExpression
        { expr = creator.createUnaryAdditiveExpression( getPosition( startToken ), AdditiveSign.PLUS, e ); }
      | MINUS e = unaryExpression
        { expr = creator.createUnaryAdditiveExpression( getPosition( startToken ), AdditiveSign.MINUS, e ); }
      | expr = unaryExpressionNotPlusMinus
    )
;

unaryExpressionNotPlusMinus returns [UnaryExpressionNotPlusMinus expr = null]
:
    expr = javaUnaryExpressionNotPlusMinus
;

javaUnaryExpressionNotPlusMinus returns [UnaryExpressionNotPlusMinus expr = null]
{
    Type t = null;
    int dim = 0;
    UnaryExpression e = null;
    Position startPosition = null;
}
:
    { startPosition = getPosition( LT(1) ); }
    (
          COMPL e = unaryExpression
          { expr = creator.createComplementExpression( startPosition, ComplementSign.BITWISE, e ); }
        | NOT e = unaryExpression
          { expr = creator.createComplementExpression( startPosition, ComplementSign.LOGICAL, e ); }
        | ( LPARENT primitiveType ( LBRACKET RBRACKET )* RPARENT unaryExpression )=>
              LPARENT t = primitiveType ( LBRACKET endToken:RBRACKET { dim++; } )* RPARENT
              e = unaryExpression
              { expr = creator.createCastExpression
                               ( startPosition
                               , dim > 0 ? creator.createArrayType( getPosition( endToken ), t, dim ) : t
                               , e
                               );
              }
        | // Have to backtrack to see if operator follows. If no operator
          // follows, it's a typecast. No semantic checking needed to parse.
          // if it _looks_ like a cast, it _is_ a cast; else it's a "(expr)"
          ( LPARENT
            qualifiedParameterizedIdentifier[false]
            ( LBRACKET RBRACKET )*
            RPARENT
            unaryExpressionNotPlusMinus
          )=>
          LPARENT
          t = qualifiedParameterizedIdentifier[false]
          ( LBRACKET arrayTypeEndToken:RBRACKET { dim++; } )*
          RPARENT
          e = unaryExpressionNotPlusMinus
          {
              expr = creator.createCastExpression
                             (  startPosition
                             , dim > 0 ? creator.createArrayType( getPosition( arrayTypeEndToken ), t, dim ) : t
                             , (UnaryExpressionNotPlusMinus)e
                             );
          }
      | expr = postfixExpression
    )
;

// qualified names, array expressions, method invocation, post inc/dec
postfixExpression returns [PostfixExpression expr = null]
:
    expr = primary
    (   incrToken:INCR
        { expr = creator.createPostAdditiveExpression( getPosition( incrToken ), expr, AdditiveSign.PLUS ); }
      | decrToken:DECR
        { expr = creator.createPostAdditiveExpression( getPosition( decrToken ), expr, AdditiveSign.MINUS ); }
    )*
;

//primary returns [Primary primary_expr = null]
//{
//}
//:
//    literal
//  | qualifiedIdentifier
//    (   ( LBRACKET  RBRACKET )+ DOT CLASS
//      | DOT ( CLASS | THIS )
//    )
//  | primitiveType ( LBRACKET  RBRACKET )* DOT CLASS
//  | VOID DOT CLASS
//  | THIS
//  | qualifiedIdentifier DOT

//  | ParenthesizedExpression
//  | ClassInstanceCreationExpression
//  | FieldAccess
//  | MethodInvocation
//  | ArrayAccess
//;

primary returns [Primary primary_expr = null]
:
    primary_expr = javaPrimary
;

// The basic element of an expression
javaPrimary returns [Primary primary_expr = null]
{
    IsPrimaryPrefixQID is_prefix_qid = new IsPrimaryPrefixQID();
    List<Identifier> idList = null;
    Primary prefix = null;
    PrimitiveType pt = null;
    int dim = 0;
    Expression expr = null;
    List<Expression> opt_args = null;
    Identifier id = null;
    TypeArguments optTypeArguments = null;
    QualifiedParameterizedIdentifier qpid = null;
    TypeArguments optConstructorTypeArguments = null;
    TypeBody optTypeBody = null;
    Token arrayTypeEndToken = null;
}
:
    // Prefix of primary expression
    (  // Literals (numeric (integral or float), boolean, null, character or string)
       prefix = literal
     | // Creation expressions
       nwt:NEW
       (  // Array creation
          ( ( primitiveType | qualifiedParameterizedIdentifier[false] ) LBRACKET )=>
          prefix = arrayCreationExpression[nwt]
        | // Unqualified class instance creation ("new <T>class_name<T>(){}")
          optConstructorTypeArguments = optTypeArguments[false, false]
          idList = identifierList
          optTypeArguments = optTypeArguments[false, false]
          LPARENT
          opt_args = optExpressionList
          argsEndToken:RPARENT
          (   arrayTypeBodyStartToken:LBRACE
              optTypeBody = typeBody[arrayTypeBodyStartToken]
              typeBodyEndToken:RBRACE
            | { optTypeBody = null; }
          )
          {
             if( createTree )
             {
                 Identifier type = (Identifier)idList.get( idList.size() - 1 );
                 idList.remove( idList.size() - 1 );
                 prefix = creator.createSimpleClassInstanceCreationExpression
                                  ( getPosition( typeBodyEndToken != null ? typeBodyEndToken : argsEndToken )
                                  , optConstructorTypeArguments
                                  , type
                                  , optTypeArguments
                                  , opt_args
                                  , optTypeBody
                                  , getPosition( nwt )
                                  , idList.isEmpty() ? null : creator.createQualifiedIdentifier( idList )
                                  );
                 idList = null;
             }
          }
       )
     | // Primaries begining with qualified identifier
       { turnOnIsPrefixQID(is_prefix_qid) }? // It is needed for guessing mode
       idList = identifierList
     | // Unqualified this access expression ("this")
       tht1:THIS
       { prefix = creator.createThisAccessExpression( getPosition( tht1 ), null ); }
     | // Unqualified super access ("super.")
       sut1:SUPER DOT
       (  // Unqualified super method invocation ("super.<T>ID()")
          ( ( typeArguments[false, false] )? identifier LPARENT )=>
          optTypeArguments = optTypeArguments[false, false]
          id = identifier
          LPARENT
          opt_args = optExpressionList
          smiEndToken:RPARENT
          { prefix = creator.createSuperMethodInvocation
                             ( getPosition( smiEndToken ), optTypeArguments, id, opt_args, getPosition( sut1 ), null ); }
        | // Unqualified super field access ("super.ID")
          id = identifier
          { prefix = creator.createSuperFieldAccess( id, getPosition( sut1 ), null ); }
       )
     | // Class literals for built-in types, their arrays and void type
       // ("builtin_type.class", "builtin_type[].class", "void.class")
       (  pt = primitiveType
          ( options { greedy = true; }: LBRACKET atEndToken:RBRACKET { dim++; } )*
          DOT clEndToken:CLASS
          {
              if(dim == 0)
              {
                  prefix = creator.createClassLiteral( getPosition( clEndToken ), pt );
              } else {
                  prefix = creator.createClassLiteral
                                   ( getPosition( clEndToken ), creator.createArrayType( getPosition( atEndToken ), pt, dim ) );
                  dim = 0;
              }
          }
        | v:VOID DOT vclEndToken:CLASS
          {
              prefix = creator.createClassLiteral( getPosition( vclEndToken ), creator.createVoidType( getPosition( v ) ) );
          }
       )
     | // Parenthesized expression ( "(expr)" )
       lpt:LPARENT expr = expression rpt:RPARENT
       { prefix = creator.createParenthesizedExpression( getPosition( lpt ), getPosition( rpt ), expr ); }
// Not allowed by JLS & javac
//     | // Unqualified parameterized name method invocation ("<T>ID()")

    ) // End of primary prefix

    // Postfix of primary expression

    (
       // Turn off the warning about non-determinism between
       // alt 3 & 4 (DOT) and exit of the loop. (See comments below.)
       options { warnWhenFollowAmbig = false; }:
       { !(LA(1) == DOT && LA(2) == SUPER && LA(3) == LPARENT) }?
    (
       // Name method invocation ("QUID()")
       { is_prefix_qid.flag }?
       LPARENT
       opt_args = optExpressionList
       snmiEndToken:RPARENT
       {
           if( createTree )
           {
               Identifier name = (Identifier)idList.get( idList.size() - 1 );
               idList.remove( idList.size() - 1 );
               prefix = creator.createQualifiedMethodInvocation
                                ( getPosition( snmiEndToken )
                                , null
                                , name
                                , opt_args
                                , idList.isEmpty() ? null : creator.createQualifiedIdentifier( idList )
                                );
               idList = null;
           }
       }
     | // Array access or class literal
       LBRACKET
       (  // Array access ("array_name[expr]" or "primary_expr[expr]")
          expr = expression aaEndToken:RBRACKET
          {
              if( prefix == null )
              {
                  // convert idList to expressionName
                  prefix = creator.createExpressionName( creator.createQualifiedIdentifier( idList ) );
                  idList = null;
              }
              prefix = creator.createArrayAccess( getPosition( aaEndToken ), prefix, expr );
          }
        | // Array reference type class literal ("type_name[].class")
          { is_prefix_qid.flag }?
                     { arrayTypeEndToken = LT(1); } RBRACKET { dim++; }
          ( LBRACKET { arrayTypeEndToken = LT(1); } RBRACKET { dim++; } )*
          DOT aclEndToken:CLASS
          {
              // TODO: check need of parameterized qualified identifiers
              // Qualified identifier --> qualified parameterized identifier
              if( createTree ) // if there were no errors
              {
                  List<ParameterizedIdentifier> pidList = listFactory.create();
                  for( int i = 0; i < idList.size(); i++ )
                  {
                      Identifier qid_i = (Identifier)idList.get(i);
                      pidList.add( creator.createParameterizedIdentifier
                                           ( creator.createIdentifier( qid_i.getPosition()
                                                                     , qid_i.getValue()
                                                                     )
                                           , null
                                           )
                                );
                  }
                  qpid = creator.createQualifiedParameterizedIdentifier( pidList );

                  prefix = creator.createClassLiteral
                                   ( getPosition( aclEndToken )
                                   , creator.createArrayType( getPosition( arrayTypeEndToken ), qpid, dim )
                                   );
                  idList = null;
              }
              dim = 0;
          }
       )
     | // Accessing (".")
       // CONFLICT: qualified super access: "primary_or_name.super."
       //           or qualified explicit constructor invocation:
       //           "primary_expr.<..>super(".
       // If ".super(" we should finish parsing the primary,
       // otherwise it should be ".super.".
       { is_prefix_qid.flag }?
       ( DOT SUPER DOT ) =>
       DOT
       // Qualified super access ("super.")
       sut2:SUPER DOT
       (  // Qualified super method invocation ("QUID.super.<T>ID()")
          ( optTypeArguments[false, false] identifier LPARENT )=>
          optTypeArguments = optTypeArguments[false, false]
          id = identifier
          LPARENT
          opt_args = optExpressionList
          qsmiEndToken:RPARENT
          {
              prefix = creator.createSuperMethodInvocation
                               ( getPosition( qsmiEndToken )
                               , optTypeArguments
                               , id
                               , opt_args
                               , getPosition( sut2 )
                               , creator.createQualifiedIdentifier( idList )
                               );
              idList = null;
          }
        | // Qualified super field access ("QUID.super.ID")
          id = identifier
          {
              prefix = creator.createSuperFieldAccess
                               ( id
                               , getPosition( sut2 )
                               , creator.createQualifiedIdentifier( idList )
                               );
              idList = null;
          }
       )
     | // method invocation ("QUID.<T>ID()") or ("primary_expr.<T>ID()")
       ( DOT optTypeArguments[false, false] identifier LPARENT )=>
       DOT
       optTypeArguments = optTypeArguments[false, false]
       id = identifier
       LPARENT
       opt_args = optExpressionList
       pmiEndToken:RPARENT
       {
           if( idList != null )
           {
               // Qualified parameterized name method invocation ("QUID.<T>ID()")
               prefix = creator.createQualifiedMethodInvocation
                                ( getPosition( pmiEndToken )
                                , optTypeArguments
                                , id
                                , opt_args
                                , creator.createQualifiedIdentifier( idList )
                                );
               idList = null;
           } else {
               // Primary method invocation ("primary_expr.<T>ID()")
               prefix = creator.createPrimaryMethodInvocation
                                ( getPosition( pmiEndToken )
                                , optTypeArguments
                                , id
                                , opt_args
                                , prefix
                                );
           }
       }
     | DOT
       (  // Not array reference type class literal ("class_name.class")
          { is_prefix_qid.flag }?
          pclEndToken:CLASS
          {
              // TODO: check need of parameterized qualified identifiers
              // identifier list --> qualified parameterized identifier
              if( createTree ) // if there were no errors
              {
                  List<ParameterizedIdentifier> pidList = listFactory.<ParameterizedIdentifier>create();
                  for( int i = 0; i < idList.size(); i++ )
                  {
                      Identifier qid_i = (Identifier)idList.get(i);
                      pidList.add( creator.createParameterizedIdentifier
                                           ( creator.createIdentifier( qid_i.getPosition()
                                                                     , qid_i.getValue()
                                                                     )
                                           , null
                                           )
                                 );
                  }
                  qpid = creator.createQualifiedParameterizedIdentifier( pidList );

                  prefix = creator.createClassLiteral( getPosition( pclEndToken ), qpid );
                  idList = null;
              }
          }
        | // Qualified this access ("name.this")
          { is_prefix_qid.flag }?
          tht2:THIS
          {
              prefix = creator.createThisAccessExpression
                               ( getPosition( tht2 )
                               , creator.createQualifiedIdentifier( idList )
                               );
              idList = null;
          }
        | // Qualified class instance creation ("primary_expr_or_ID.new <T>ID<T>(){}")
          NEW
          optConstructorTypeArguments = optTypeArguments[false, false]
          id = identifier
          optTypeArguments = optTypeArguments[false, false]
          LPARENT
          opt_args = optExpressionList
          primArgsEndToken:RPARENT
          (   classTypeBodyStartToken:LBRACE
              optTypeBody = typeBody[classTypeBodyStartToken]
              primTypeBodyEndToken:RBRACE
            | { optTypeBody = null; }
          )
          {
              /* If idList != null then it is name qualified class instance creation
               * ("name.new <T>id<T>(){}") else it is primary qualified
               * class instance creation ("primary_expr.new <T>id<T>(){}")
               */
              if( idList != null )
              {
                  prefix = creator.createExpressionName( creator.createQualifiedIdentifier( idList ) );
                  idList = null;
              }
              prefix = creator.createPrimaryClassInstanceCreationExpression
                               ( getPosition( primTypeBodyEndToken != null ? primTypeBodyEndToken : primArgsEndToken )
                               , optConstructorTypeArguments
                               , id
                               , optTypeArguments
                               , opt_args
                               , optTypeBody
                               , prefix
                               );
          }
        | // Primary field access ("primary_expr.ID")
          id = identifier
          { prefix = creator.createPrimaryFieldAccess(id, prefix); }
       )
    )
    { turnOffIsPrefixQID( is_prefix_qid ) }?
    )*// End of primary postfix
    {
        if( idList != null )
        {
            primary_expr = creator.createExpressionName( creator.createQualifiedIdentifier( idList ) );
        } else {
            primary_expr = prefix;
        }
    }
;

arrayCreationExpression[Token newToken] returns [ArrayCreationExpression expr = null]
{
    Type t = null;
    int dim = 0;
    Expression e = null;
    List<Expression> exprs = null;
    ArrayInitializer array_init = null;
    Token endToken = null;
}
:
    ( t = primitiveType | t = qualifiedParameterizedIdentifier[false] )
    (  ( LBRACKET { dim++; } RBRACKET )+ array_init = arrayInitializer
       { expr = creator.createInitArrayCreationExpression( getPosition( newToken ), t, dim, array_init ); }
     | // CONFLICT:
       // New array expression is a primary expression which can be
       // followed by an array index reference. The generated code
       // will stay in this loop as long as it sees "["
       // (it is a proper behavior)
       { exprs = listFactory.<Expression>create(); }
       ( options { greedy = true; }:
         LBRACKET e = expression { endToken = LT(1); } RBRACKET { listFactory.add(exprs, e); }
       )+
       ( options { greedy = true; }: LBRACKET { dim++; endToken = LT(1); } RBRACKET )*
       { expr = creator.createExprArrayCreationExpression( getPosition( newToken ), t, dim, exprs, getPosition( endToken ) ); }
    )
;

// Class literals "<type>.class"
classLiteral returns [ClassLiteral cl = null]
{
     OptionalType type = null;
}
:
    type = optionalType[false]
    DOT endToken:CLASS
    { cl = creator.createClassLiteral( getPosition( endToken ), type ); }
;

// Literals
literal returns [Literal lit = null]
{
    Token token = null;
    String txt = null;
    Position position = null;
}
:
    { token = LT(1); txt = token.getText(); position = getPosition( token ); }
    (   (  TRUE | FALSE )      { lit = creator.createBooleanLiteral( position, txt ); }
      | INTEGER_LITERAL        { lit = creator.createIntegerLiteral( position, txt ); }
      | FLOATING_POINT_LITERAL { lit = creator.createFloatingPointLiteral( position, txt ); }
      | CHARACTER_LITERAL      { lit = creator.createCharacterLiteral( position, txt ); }
      | STRING_LITERAL         { lit = creator.createStringLiteral( position, txt ); }
      | NULL                   { lit = creator.createNullLiteral( position, txt ); }
    )
;

//=====  Identifiers  ================================================

identifierList returns [List<Identifier> idList = null]
{
    Identifier id = null;
    boolean cTree = createTree;
}
:
    { createTree( true ); idList = listFactory.create(); createTree( cTree ); }
    id = identifier
    {
    	cTree = createTree;
        listFactory.add( idList, id );
        createTree( true );
    }
    ( options { greedy=true; }: // match as many as possible
      DOT id = identifier
      {
        listFactory.add( idList, id );
        createTree( true );
      }
    )*
    {
        createTree( cTree );
    }
;

// A (possibly-qualified) java identifier. We start with the first IDENT
// and expand its name by adding dots and following IDENTS
qualifiedIdentifier returns [QualifiedIdentifier qid = null]
{
    List<Identifier> idList = null;
}
:
    idList = identifierList
    {
        qid = creator.createQualifiedIdentifier( idList );
    }
;

// A single (non-qualified and non-parameterized) java identifier.
identifier returns [Identifier ident = null]
:
    t_id: IDENTIFIER
    {
        ident = creator.createIdentifier( getPosition( t_id ), t_id.getText() );
    }
;
